#!/usr/bin/env python
"""
Delta values from downloaded IsoGSM2 data.

Calculates delta values either from the original yearly IsoGSM2 files or
from the (multi-year) files produced get_isogsm2.


--------------------------------------------------------
usage: delta_isogsm2.py [-h] [-o output_file] [-t timestep]
                        [isogsm_output_file]

Calculate delta values from IsoGSM2 output.

positional arguments:
  isogsm_output_file    Output file of IsoGSM2.

optional arguments:
  -h, --help            show this help message and exit
  -o output_file, --output output_file
                        Output file name (default: ifile-delta.csv).
  -t timestep, --timestep timestep
                        Output timestep in seconds (default: IsoGSM2 timestep
                        of 6 hours = 21600).


Example
-------
# Hesse
python get_isogsm2.py -l 48.6742166667,7.06461666667
python delta_isogsm2.py -o delta_hesse_isogsm2_1979-2017.dat \
    lat48.571lon7.500_isogsm2_6hrly_1979-2017.dat

for i in $(ls lat*_isogsm2_*.dat) ; do python delta_isogsm2.py ${i} ; done


History
-------
Written,  Matthias Cuntz, Oct 2018
Modified, Matthias Cuntz, Nov 2018
              - make it callable function as well as script.
          Matthias Cuntz, Nov 2018
              - flak8 compatible
              - missing value of IsoGSM2 is 9.999e20
                -> precip, shm > 1e+15 = -1000
         Matthias Cuntz, Sep 2021
              - no error but only message when called without IsoGSM2 output
         Matthias Cuntz, Sep 2022
              - add timestep

"""
from datetime import datetime
import numpy as np
import netCDF4 as nc


# -------------------------------------------------------------------------
# Command line arguments
#

def delta_isogsm2(ifile, output=None, timestep=21600):
    '''
    Delta values from downloaded IsoGSM2 data

    Calculate delta values either from the original yearly IsoGSM2 files
    or from the (multi-year) files produced by get_isogsm2.
    Values can get interpolated to a desired output timestep.

    The script expects that the year(s) is(are) given at the end of the
    filename before the suffix, separated by _ from the rest of the
    filename, because there is no date info in the IsoGSM2 files.
    A year range is given by -, e.g.
        x001y192_isogsm2_6hrly_1999.dat
    or
        lat48.571lon7.500_isogsm2_6hrly_1979-2017.dat

    Parameters
    ----------
    ifile : string
        filename of IsoGSM2 output
    output : string, optional
        output filename. Default: ifile-delta.csv
    timestep : int, optional
        output time step in seconds. Default: 21600 (= 6 hours)
        IsoGSM2 output is 6-hourly. Interpolation will be linear for
        everything but precipitation. The latter will be distributed
        equally within the 6 hours.

    Returns
    -------
    filename of output file

    Notes
    -----
    The output file has the following columns:
       Date - timestep in the middle of the time period
       precip (mm) - cumulative precipitation during time step
       d18oprecip (permil) - delta-18O of precipitation
       d2hprecip (permil) - delta-2H of precipitation
       sh2m (kg/kg) - specific humidity at 2 m
       d18ovap (permil) - delta-18O of atmospheric vapour
       d2hvap (permil) - delta-2H of atmospheric vapour
       tair (K) - air temperature
       rh (percent) - relative humidity
       p (Pa) - surface pressure

    All delta values are relative to V-SMOW.

    Examples
    --------
    # All available IsoGSM2 output for FR-Hes
    latlon = '48.6742166667,7.06461666667'
    fisogsm2 = get_isogsm2(latlon)
    # delta values at half-hourly resolution
    disogsm2 = delta_isogsm2(fisogsm2, timestep=30*60)

    '''
    if timestep > 21600:
        raise ValueError('Timestep must be less or equal than 6 hours,'
                         ' i.e. 21600 seconds')
    if (86400 % timestep) > 0:
        raise ValueError(f'Seconds per day (86400) must be a multiple'
                         f' of timestep: {timestep}.')
    if (21600 % timestep) > 0:
        raise ValueError(f'IsoGSM2 timestep, 6 hours = 21600 seconds, must be'
                         f' a multiple of timestep: {timestep}.')

    if (output is None):
        output = ifile[0:ifile.rfind('.')] + '-delta.csv'

    # -------------------------------------------------------------------------
    # Data
    #

    # Lon= 7.5 Lat= 48.571
    # PWAT PWAT18O PWATHDO PRATE PRATE18O PRATEHDO CPRAT CPRAT18O CPRATHDO
    # LHTFL LHTFL18O LHTFLHDO SPFH2m SPFH2m18O SPFH2mHDO TMP2m RH2m PRESsfc
    # HGT500
    # 0.6678350E+01 0.6431830E+01 0.4832670E+01 0.3923000E-04 0.3868800E-04
    # 0.3539200E-04 0.0000000E+00 0.0000000E+00 0.0000000E+00 -0.1431808E+02
    # -0.1416832E+02 -0.1319024E+02 0.2095527E-02 0.2035712E-02 0.1645568E-02
    # 0.2643900E+03 0.1000000E+03 0.9299900E+05 0.5445400E+04

    # print('Read ', ifile)
    # Determine years from filename
    ff1 = ifile[0:ifile.rfind('.')]
    yrs = ff1[ff1.rfind('_') + 1:]
    if '-' in yrs:
        try:
            yr1, yr2 = [ int(yy) for yy in yrs.split('-') ]
        except:
            raise ValueError('Could not determine years from file name. (1)')
    else:
        try:
            yr1 = yr2 = int(yrs)
        except:
            raise ValueError('Could not determine years from file name. (2)')
    # nyr = yr2 - yr1 + 1

    # Read file
    ff = open(ifile, 'r')
    lon, lat = [ float(ll) for ll in ff.readline().strip().split()[1::2] ]
    head = ff.readline().split()
    ff.close()
    dat = np.loadtxt(ifile, skiprows=2)
    ndat = dat.shape[0]

    # Make dates
    units = 'days since ' + str(yr1) + '-01-01 00:00:00'
    # date0 is central in time step
    gtimestep = 21600
    second = 0
    minute = 0
    hour = 3
    date0 = nc.date2num(datetime(yr1, 1, 1, hour, minute, second),
                        units)
    # fractional days since 01.01.Year1
    fdate = gtimestep / 86400.
    jdate = date0 + np.arange(ndat) * fdate
    # datetime objects
    ddate = nc.num2date(jdate, units)
    # ascii date (eng)
    adate = [ d.isoformat(sep=' ') for d in ddate ]

    # Get deltas
    vprecip = 'PRATE'
    vvap    = 'SPFH2m'
    vtair   = 'TMP2m'
    vrh     = 'RH2m'
    vp      = 'PRESsfc'

    precip     = dat[:, head.index(vprecip)]
    # permil
    d18oprecip = (dat[:, head.index(vprecip + '18O')] / precip - 1.) * 1000.
    d2hprecip  = (dat[:, head.index(vprecip + 'HDO')] / precip - 1.) * 1000.
    # current undef of IsoGSM is 9.999e20
    ii = np.where((precip < 1.e-15) | (precip > 1.e+15))[0]
    if ii.size > 0:
        d18oprecip[ii] = -1000.
        d2hprecip[ii]  = -1000.
    sh2m    = dat[:, head.index(vvap)]  # specific humidity (kg/kg)
    d18ovap = (dat[:, head.index(vvap + '18O')] / sh2m - 1.) * 1000.
    d2hvap  = (dat[:, head.index(vvap + 'HDO')] / sh2m - 1.) * 1000.
    ii = np.where((sh2m < 1.e-15) | (sh2m > 1.e+15))[0]
    if ii.size > 0:
        d18ovap[ii] = -1000.
        d2hvap[ii]  = -1000.
    tair = dat[:, head.index(vtair)]  # K
    rh   = dat[:, head.index(vrh)]    # %
    p    = dat[:, head.index(vp)]     # Pa

    # interpolate to timestep
    if timestep == 21600:
        odate = adate
    else:
        # date0 is central in time step
        ihoy = int(timestep // 2)
        second = ihoy % 60
        ihoy   = ihoy // 60
        minute = ihoy % 60
        ihoy   = ihoy // 60
        hour = ihoy % 24
        ihoy = ihoy // 24
        # hour0 = np.rint(fdate * 24.)
        odate0 = nc.date2num(datetime(yr1, 1, 1, hour, minute, second),
                             units)
        fodate = timestep / 86400.
        nodat = (ndat // 4) * (86400 // timestep)
        jodate = odate0 + np.arange(nodat) * fodate
        dodate = nc.num2date(jodate, units)
        odate = [ d.isoformat(sep=' ') for d in dodate ]
        # continues variables
        sh2m = np.interp(jodate, jdate, sh2m)
        d18ovap = np.interp(jodate, jdate, d18ovap)
        d2hvap = np.interp(jodate, jdate, d2hvap)
        tair = np.interp(jodate, jdate, tair)
        rh = np.interp(jodate, jdate, rh)
        p = np.interp(jodate, jdate, p)
        # rain isotopes
        # interpolate over gaps and set to undef later
        ii = np.where(d18oprecip != -1000.)[0]
        if ii.size > 0.:
            d18oprecip = np.interp(jodate, jdate[ii], d18oprecip[ii])
        else:
            d18oprecip = np.full(nodat, -1000.)
        ii = np.where(d2hprecip != -1000.)[0]
        if ii.size > 0.:
            d2hprecip = np.interp(jodate, jdate[ii], d2hprecip[ii])
        else:
            d2hprecip = np.full(nodat, -1000.)
        # rain
        # distribute equally over 6 hours
        precip = precip * (timestep / gtimestep)
        oprecip = np.full(nodat, 0.)
        n = gtimestep // timestep
        for i in range(ndat):
            oprecip[i * n:(i + 1) * n] = precip[i]
        precip = oprecip
        ii = np.where(precip == 0.)[0]
        if ii.size > 0:
            d18oprecip[ii] = -1000.
            d2hprecip[ii] = -1000.

    # Write file
    # print('Write ', output)
    with open(output, 'w') as ff:
        # print('Lon={:.2f},Lat={:.3f}'.format(lon,lat), file=ff)
        astr = ('Date'
                ',precip (mm),d18oprecip (permil),d2hprecip (permil)'
                ',sh2m (kg/kg),d18ovap (permil),d2hvap (permil)'
                ',tair (K),rh (percent),p (Pa)')
        print(astr, file=ff)
        for i in range(tair.size):
            astr = odate[i]
            astr = astr + ',{:.6e},{:.4f},{:.4f}'.format(
                precip[i], d18oprecip[i], d2hprecip[i])
            astr = astr + ',{:.6e},{:.4f},{:.4f}'.format(
                sh2m[i], d18ovap[i], d2hvap[i])
            astr = astr + ',{:.2f},{:.4f},{:.0f}'.format(
                tair[i], rh[i], p[i])
            print(astr, file=ff)

    return output


def main(iargs):

    import argparse

    timestep = 21600  # 6 * 3600
    ofile  = None
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description='''Calculate delta values from IsoGSM2 output.''')
    parser.add_argument('-o', '--output', action='store',
                        default=ofile, dest='ofile', metavar='output_file',
                        help='Output file name (default: ifile-delta.csv).')
    parser.add_argument('-t', '--timestep', action='store',
                        default=timestep, dest='timestep',
                        metavar='timestep',
                        help=('Output timestep in seconds (default:'
                              ' IsoGSM2 timestep of 6 hours = 21600).'))
    parser.add_argument('ifile', nargs='?', default=None,
                        metavar='isogsm_output_file',
                        help='Output file of IsoGSM2.')

    args = parser.parse_args(iargs)
    ofile    = args.ofile
    timestep = int(args.timestep)
    ifile    = args.ifile

    del parser, args

    # Check input
    if (ifile is None):
        print('No IsoGSM2 output file given.')
        import sys
        sys.exit()

    import time as ptime
    t1 = ptime.time()

    dfile = delta_isogsm2(ifile, output=ofile, timestep=timestep)
    print('File: ', dfile)

    t2    = ptime.time()
    strin = ('[m]: {:.1f}'.format((t2 - t1) / 60.) if (t2 - t1) > 60. else
             '[s]: {:d}'.format(int(t2 - t1)))
    print('Time elapsed', strin)



#!/usr/bin/env python
import numpy as np
from pyjams.kernel_regression import kernel_regression


__all__ = ['fill_nonfinite']


def _nan_helper(y, nan=False, inf=False, undef=None):
    """
    Helper to handle indices and logical indices of NaNs, Infs or undefs.


    Definition
    ----------
    def _nan_helper(y, nan=False, inf=False, undef=None):


    Input
    -----
    y        1d numpy array with possible missing values


    Optional Input
    --------------
    At least one of the following has to be given
    nan        if True, check only for NaN and not Inf.
    inf        if True, check only for Inf and not NaN.
    undef      if given then check for undef value rather than NaN and Inf.


    Output
    ------
    ind      logical indices of missing values
    find     function, with signature indices = find(ind),
             to convert logical indices of NaNs to 'equivalent' indices


    Examples
    --------
    >>> # linear interpolation of NaNs
    >>> y       = np.array([1, np.nan, 3])
    >>> nans, z = _nan_helper(y, nan=True)
    >>> y[nans] = np.interp(z(nans), z(~nans), y[~nans])


    History
    -------
    Written,  Matthias Cuntz, Jul 2013 - modified from
        http://stackoverflow.com/questions/6518811/interpolate-nan-values-in-a-numpy-array
    Modified, Matthias Cuntz, Apr 2014 - assert
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    assert not ((not nan) & (not inf) & (undef is None)), (
        'at least one of nan, inf or undef has to be given.')

    out = np.zeros(y.shape, dtype=bool)
    if nan:
        out = out | np.isnan(y)
    if inf:
        out = out | np.isinf(y)
    if undef is not None:
        out = out | (y == undef)

    return out, lambda ind: ind.nonzero()[0]


def fill_nonfinite(xin, yin=None, nan=None, inf=None, undef=None,
                   method='interp', **kwargs):
    """
    Fill missing values by interpolation.

    Checks for NaN and Inf in xin and interpolates along column (1st dim).
    If yin is None, then interpolation distance is determined by index number.
    If undef=number is given, it will be interpolated as well.
    If one of nan of inf is given, then only that will be checked,
    i.e. no keyword or only undef given: nan=True, inf=True.
         only nan and undef given: nan=True, inf=False.


    Definition
    ----------
    def fill_nonfinite(xin, yin=None, nan=None, inf=None, undef=None,
    method='interp', **kwargs):


    Input
    -----
    xin     ND array


    Optional Input
    --------------
    yin        ND array (default: None)
               If not given, then y=xin and x=np.arange(len(xin[:,0]))
    nan        if True, interpolate NaN (default: None, see above).
    inf        if True, interpolate Inf (default: None, see above).
    undef      if given, interpolate xin==undef (default: None).
    method     Interpolation method (default: 'interp'):
               'interp': linear interpolation with np.interp.
               'linear': same as 'interp'
               'kernel_regression': non-linear interpolation with
               jams.kernel_regression.
    **kwargs   All other keyword arguments will be passed to
               interpolation routine.


    Output
    ------
    xin filled at NaN and Inf values.


    Examples
    --------
    >>> from autostring import astr

    >>> a = np.arange(3)+1.
    >>> print(astr(fill_nonfinite(a),1,pp=True))
    ['1.0' '2.0' '3.0']

    >>> a[1] = np.nan
    >>> print(astr(fill_nonfinite(a),1,pp=True))
    ['1.0' '2.0' '3.0']

    >>> a[1] = np.inf
    >>> print(astr(fill_nonfinite(a,inf=True),1,pp=True))
    ['1.0' '2.0' '3.0']

    >>> a = np.arange(3)+1.
    >>> print(astr(fill_nonfinite(a,undef=2),1,pp=True))
    ['1.0' '2.0' '3.0']

    >>> print(astr(fill_nonfinite(a,undef=2,method='Kernel_regression'),1,pp=True))
    ['1.0' '2.0' '3.0']
    >>> print(astr(fill_nonfinite(a,undef=2,method='Kernel_regression',silverman=True),1,pp=True))
    ['1.0' '2.0' '3.0']

    >>> a = np.arange(10)+1.
    >>> x = a[:]
    >>> x[1::3] -= 0.5
    >>> print(astr(fill_nonfinite(x,a,undef=2),1,pp=True))
    [' 1.0' ' 1.5' ' 3.0' ' 4.0' ' 4.5' ' 6.0' ' 7.0' ' 7.5' ' 9.0' '10.0']
    >>> print(astr(fill_nonfinite(x,a,undef=2,method='Kernel_regression',silverman=True),1,pp=True))
    [' 1.0' ' 1.5' ' 3.0' ' 4.0' ' 4.5' ' 6.0' ' 7.0' ' 7.5' ' 9.0' '10.0']


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2013-2021 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Matthias Cuntz, Jul 2013
    Modified, Matthias Cuntz, Jan 2018 - yin
              Matthias Cuntz, Jan 2018 - added kernel_regression and **kwargs
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    #
    # yin
    if yin is not None:
        xx = xin
        yy = yin
    else:
        xx = None
        yy = xin
    #
    # Assure ND-array
    isone = False
    if np.ndim(yy) == 1:
        isone = True
        yy = yy[:, np.newaxis]
        ny = 1
    else:
        ny = yy.shape[1]
    if xx is None:
        nx = 0
    else:
        if np.ndim(xx) == 1:
            xx = xx[:, np.newaxis]
            nx = 1
        else:
            nx = xx.shape[1]
            assert nx == ny, 'x and y ND-arrays must have same dimensions if x is not 1D-array.'
    #
    # Check options - if none given then nan and inf are True
    inan = False
    iinf = False
    if (nan is None) & (inf is None):
        inan = True
        iinf = True
    else:
        if (nan is not None):
            inan = nan
        if (inf is not None):
            iinf = inf

    # interpolate along each column
    yout = yy.astype(float)  # np.interp always returns float
    for i in range(ny):
        yy1 = yout[:, i]
        if xx is not None:
            xx1 = xx[:, min(i, nx-1)].astype(float)
        nans, ind = _nan_helper(yy1, nan=inan, inf=iinf, undef=undef)
        if np.sum(nans) > 0:
            if (method.lower() == 'interp') or (method.lower() == 'linear'):
                if xx is None:
                    yy1[nans] = np.interp(ind(nans), ind(~nans), yy1[~nans],
                                          **kwargs)
                else:
                    yy1[nans] = np.interp(xx1[nans], xx1[~nans], yy1[~nans],
                                          **kwargs)
            elif method.lower() == 'kernel_regression':
                if xx is None:
                    yy1[nans] = kernel_regression(ind(~nans), yy1[~nans],
                                                  xout=ind(nans), **kwargs)
                else:
                    yy1[nans] = kernel_regression(xx1[~nans], yy1[~nans],
                                                  xout=xx1[nans], **kwargs)
            else:
                raise ValueError('Interpolation method unknown:', method)
            yout[:, i] = yy1

    if isone:
        return np.squeeze(yout)
    else:
        return yout


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # from autostring import astr
    # a = np.arange(3)+1.
    # print(astr(fill_nonfinite(a),1,pp=True))
    # #['1.0' '2.0' '3.0']
    # a[1] = np.nan
    # print(astr(fill_nonfinite(a),1,pp=True))
    # #['1.0' '2.0' '3.0']
    # a[1] = np.inf
    # print(astr(fill_nonfinite(a,inf=True),1,pp=True))
    # #['1.0' '2.0' '3.0']
    # a = np.arange(3)+1.
    # print(astr(fill_nonfinite(a,undef=2),1,pp=True))
    # #['1.0' '2.0' '3.0']
    # print(astr(fill_nonfinite(a,undef=2,method='Kernel_regression'),1,pp=True))
    # #['1.0' '2.0' '3.0']
    # print(astr(fill_nonfinite(a,undef=2,method='Kernel_regression',silverman=True),1,pp=True))
    # #['1.0' '2.0' '3.0']
    # a = np.arange(10)+1.
    # x = a[:]
    # x[1::3] -= 0.5
    # print(astr(fill_nonfinite(x,a,undef=2),1,pp=True))
    # #[' 1.0' ' 1.5' ' 3.0' ' 4.0' ' 4.5' ' 6.0' ' 7.0' ' 7.5' ' 9.0' '10.0']
    # print(astr(fill_nonfinite(x,a,undef=2,method='Kernel_regression',silverman=True),1,pp=True))
    # #[' 1.0' ' 1.5' ' 3.0' ' 4.0' ' 4.5' ' 6.0' ' 7.0' ' 7.5' ' 9.0' '10.0']



#!/usr/bin/env python
import numpy as np


def fwrite(fname, arr, header=None, precision='10.0', delimiter=' '):
    """
        Write numbers of 2D-array to a file.

        A header can be given as optional as well as a precision.


        Definition
        ----------
        def fwrite(fname, data, header=None, precision='10.0'):


        Input
        -----
        fname        target file name
        arr          2d numpy array to write


        Optional Input Parameters
        -------------------------
        header       list of header elements: a header element is a list of two strings:
                     the first entry is the header argument, the second the value
        precision    floating point precision of array to write
        delimiter    delimiter to separate values, default ' '


        Examples
        --------
        >>> from jams import fread
        >>> # Clean up doctest
        >>> filename = 'fwrite.test'
        >>> header = [['Description', 'testing'], ['author', 'ST']]
        >>> data = np.arange(10).reshape(2, 5)
        >>> fwrite(filename, data, header=header)

        >>> fread(filename, nc=2, skip=2, header=True)
        [['Description', 'testing'], ['author', 'ST']]
        >>> fread(filename, skip=2)
        array([[0., 1., 2., 3., 4.],
               [5., 6., 7., 8., 9.]])

        >>> import os
        >>> os.remove(filename)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016-2017 Stephan Thober

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  ST, Feb 2016
        Modified, ST, Aug 2017 - delimiter
    """
    if not type(arr) is np.ndarray:
        raise ValueError('function fwrite: argument arr must be numpy.ndarray')
    fo = open(fname, 'w')
    if not header is None:
        # write header
        for ll in np.arange(len(header)):
            write_str = str(header[ll][0]) + ' ' + str(header[ll][1]) + '\n'
            fo.write(write_str)
    # write arr
    for ll in np.arange(arr.shape[0]):
        # format is mRM compatible
        write_str = delimiter.join(['{:' + precision + 'f}'] * arr.shape[1]).format(*arr[ll, :]) + '\n'
        fo.write(write_str)
    fo.close()


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def readhdf5(fName, var='', reform=False, squeeze=False, variables=False,
             attributes=False, fileattributes=False, sort=False):
    """
        Get variables or print information of hdf5 file.


        Definition
        ----------
        def readhdf5(fName, var='', reform=False, squeeze=False, variables=False,
                     attributes=False, fileattributes=False, sort=False):

        Input
        -----
        fName            hdf5 file name


        Optional Input Parameters
        -------------------------
        var              name of variable in hdf5 file


        Options
        -------
        reform           if output is array then squeeze(array)
        squeeze          same as reform
        variables        get list of variables in hdf5 file
        attributes       get dictionary of all attributes of specific variable
        fileattributes   get dictionary of all attributes of the file
        sort             sort variable names.
        quiet            quietly return None if error occurs


        Output
        ------
        Either variable array or information such as list of all variables in hdf5 file.


        Examples
        --------
        >>> a = readhdf5('test_readhdf5.hdf5', fileattributes=True)
        >>> print(a['NB_PARAMETERS'])
        9

        >>> print([str(i) for i in readhdf5('test_readhdf5.hdf5', variables=True)])
        ['chs']

        >>> a = readhdf5('test_readhdf5.hdf5', var='chs', attributes=True)
        >>> print([ str(i) for i in sorted(a)])
        ['Double', 'Inttest', 'LLLLLL', 'What the hell']
        >>> print(a['Double'])
        [1.1]

        >>> from autostring import astr
        >>> print(astr(readhdf5('test_readhdf5.hdf5', var='chs'),3,pp=True))
        [['  1.000' '  2.000' '  3.000' '  3.000' '  2.000']
         ['  1.000' ' 23.000' '254.000' '  5.000' '  4.654']
         ['654.654' ' 54.540' '546.540' '564.546' '  5.500']
         ['  1.100' '  2.200' '  0.000' '  3.300' '  4.400']
         ['  0.000' '  1.000' '  2.000' '  3.000' '  0.000']]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2013 Matthias Zink, Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MZ, Jun 2012
        Modified, MC, Feb 2013 - ported to Python 3
                  MC, Oct 2013 - hdf5read
    """
    try:
        import h5py as hdf5
    except:
        raise Error('No HDF5 support available, i.e. h5py')
    # Open hdf5 file
    try:
        f = hdf5.File(fName, 'r')
    except IOError:
        raise IOError('Cannot open file: '+fName)
    # Get attributes of the file
    if fileattributes:
        attrs = dict()
        attr = list(f.attrs.keys())
        for a in attr:
            attrs[a] = f.attrs.get(a)
        f.close()
        return attrs
    # Variables
    vars = list(f.keys())
    nvars = len(vars)
    # Sort and get sort indices
    if sort:
      svars = sorted(vars)
      ivars = list()
      for v in svars:
        ivars.append(vars.index(v))
    if variables:
      f.close()
      if sort:
        return svars
      else:
        return vars
    # Get attributes of variables
    if attributes:
      if var not in vars:
          f.close()
          raise ValueError('variable '+var+' not in file '+fname)
      attrs = dict()
      attr = list(f[var].attrs.keys())
      for a in attr:
        attrs[a] = f[var].attrs.get(a)
      f.close()
      return attrs
    # Get variable
    if var == '':
        f.close()
        raise ValueError('Variable name has to be given.')
    if var != '':
      if var not in vars:
          f.close()
          raise ValueError('variable '+var+' not in file '+fname)
      try:
          arr = f[var][:]
      except IOError:
          f.close()
          raise IOError('Cannot read variable '+var+' in file '+fname)
      if reform or squeeze:
        f.close()
        return arr.squeeze()
      else:
        f.close()
        return arr


def hdf5read(*args, **kwargs):
    """
        Wrapper for readhdf5.
        def readhdf5(fName, var='', reform=False, squeeze=False, variables=False,
                     attributes=False, fileattributes=False, sort=False):


        Examples
        --------
        >>> a = hdf5read('test_readhdf5.hdf5', fileattributes=True)
        >>> print(a['NB_PARAMETERS'])
        9

        >>> print([str(i) for i in hdf5read('test_readhdf5.hdf5', variables=True)])
        ['chs']

        >>> a = hdf5read('test_readhdf5.hdf5', var='chs', attributes=True)
        >>> print([ str(i) for i in sorted(a)])
        ['Double', 'Inttest', 'LLLLLL', 'What the hell']
        >>> print(a['Double'])
        [1.1]

        >>> from autostring import astr
        >>> print(astr(hdf5read('test_readhdf5.hdf5', var='chs'),3,pp=True))
        [['  1.000' '  2.000' '  3.000' '  3.000' '  2.000']
         ['  1.000' ' 23.000' '254.000' '  5.000' '  4.654']
         ['654.654' ' 54.540' '546.540' '564.546' '  5.500']
         ['  1.100' '  2.200' '  0.000' '  3.300' '  4.400']
         ['  0.000' '  1.000' '  2.000' '  3.000' '  0.000']]
    """
    return readhdf5(*args, **kwargs)


if __name__ == '__main__':
    import doctest
    try:
        import h5py as hdf5
    except ImportError:
        raise ImportError('No HDF5 support available, i.e. h5py')
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # var = readhdf5('test_readhdf5.hdf5', variables=True)
    # print var

    # varAtt = readhdf5('test_readhdf5.hdf5', var='EM_BB', attributes=True)
    # print varAtt
    # #{u'CAL_SLOPE': 999.0, u'PRODUCT': 'EM_BB', u'PRODUCT_ID': 999, u'CAL_OFFSET': 999.0, u'N_LINES': 651, u'N_COLS': 1701, u'SCALING_FACTOR': 10000.0, u'NB_BYTES': 2, u'OFFSET': 0.0, u'UNITS': 'Dimensionless', u'MISSING_VALUE': -8000, u'CLASS': 'Data'}
    # varAtt['SCALING_FACTOR']
    # #1000.0
    # data = readhdf5('test_readhdf5.hdf5', var='EM_BB')
    # print data
    # #array([[-8000, -8000, -8000, ..., -8000, -8000, -8000],
    # #       [-8000, -8000, -8000, ..., -8000, -8000, -8000],
    # #       [-8000, -8000, -8000, ..., -8000, -8000, -8000],
    # #       ...,
    # #        [-8000, -8000, -8000, ..., -8000, -8000, -8000],
    # #       [-8000, -8000, -8000, ..., -8000, -8000, -8000],
    # #       [-8000, -8000, -8000, ..., -8000, -8000, -8000]], dtype=int16)




#!/usr/bin/env python


def smax(x, y, eta=1. - 1.e-5):
    """
        Calculate smooth maximum of two numbers.


        Definition
        ----------
        def around(x, y, eta=1.):


        Input
        -----
        x        ND array of floats
        y        ND array of floats


        Optional Input
        --------------
        eta        Shape of smoothing function (default: 1._dp)


        Output
        ------
        Smoothed maximum


        Restrictions
        ------------
        If eta is set to 1, then this function becomes the standard maximum function.


        Examples
        --------
        >>> from jams.autostring import astr
        >>> print(astr(smax(3.5967, 3.6),3,pp=True))
            3.610


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Stephan Thober

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  ST, Mar 2016
        Modified, 
    """
    #
    if (eta > 1.):
        raise ValueError('eta is {:3.4f}, but cannot be larger than 1.'.format(eta))
    #
    out = (x + y + ((x + y)**2. - 4 * eta * x * y)**0.5) / (2. * eta)
    return out


def smin(x, y, eta=1. - 1.e-5):
    """
        Calculate smooth minimum of two numbers.


        Definition
        ----------
        def around(x, y, eta=1. - 1.e-5):


        Input
        -----
        x        ND array of floats
        y        ND array of floats


        Optional Input
        --------------
        eta        Shape of smoothing function (default: 1._dp)


        Output
        ------
        Smoothed minimum


        Restrictions
        ------------
        If eta is set to 1, then this function becomes the standard minimum function.


        Examples
        --------
        >>> from jams.autostring import astr
        >>> print(astr(smin(3.5967, 3.6),3,pp=True))
            3.587


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Stephan Thober

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  ST, Mar 2016
        Modified, 
    """
    #
    if (eta > 1.):
        raise ValueError('eta is {:3.4f}, but cannot be larger than 1.'.format(eta))
    #
    out = (x + y - ((x + y)**2. - 4 * eta * x * y)**0.5) / (2. * eta)
    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
#
# This script reads a DEM from a netcdf file and calculates the flow direction
# following ArcGis convention:
#
# flow direction is assumed like this
#           64             y-axis
#       32      128          |
#   16      -1       1       |
#        8       2          \|/
#            4               V
#      x-axis ------>
# ORIGIN is in the upper left corner
#
# sinks are marked by -1
#
# Nomenclature:
#
# co - channel orders
# fd - flow direction
# fa - flow accumulation
# yy - row index
# xx - column index
#
# author:  Stephan Thober, David Schaefer
# created: 01.07.2015
import numpy as np

# global variables for calculating upstream cells
# yy_offset, xx_offset = np.meshgrid(np.arange(-1, 2), np.arange(-1, 2))
yy_offset, xx_offset = np.meshgrid(np.arange(1, -2, -1), np.arange(-1, 2))
yy_offset = yy_offset.ravel()  # yy_offset for neighboring cells
xx_offset = xx_offset.ravel()  # xx_offset for neighboring cells
# local flow direction sorted in the order of yy_offset and xx_offset
local_flow_direction = np.array([8, 16, 32, 4, -1, 64, 2, 1, 128])
# same as local_flow_direction but reverted. This means these are the
# flow directions of neighboring cells flowing into the local cell.
inflow_direction = np.array([128, 1, 2, 64, -9999, 4, 32, 16, 8])


def cal_fdir(locs, fdir, factor):
    # flow direction is assumed like this
    #           64             y-axis
    #       32      128          |
    #   16      -1       1       |
    #        8       2          \|/
    #            4               V
    #      x-axis ------>
    #
    # ORIGIN is in the upper left corner -> y axis increases to the bottom
    # loc = [nrow, ncol]
    nloc = locs[0].shape[0]
    fds = np.zeros((nloc))
    for kk in range(nloc):
        # loop over all locations having maximum flow direction
        loc = [locs[0][kk], locs[1][kk]]
        fd = fdir[loc[0], loc[1]]
        if fd == 2:
            if loc[1] + 1 < factor:
                fd = 4
            elif loc[0] + 1 < factor:
                fd = 1
        elif fd == 8:
            if loc[1] > 0:
                fd = 4
            elif loc[0] + 1 < factor:
                fd = 16
        elif fd == 32:
            if loc[1] > 0:
                fd = 64
            elif loc[0] > 0:
                fd = 16
        elif fd == 128:
            if loc[1] + 1 < factor:
                fd = 64
            elif loc[0] > 0:
                fd = 1
        fds[kk] = fd
    return fds


def upscale_fdir(sn, factor, print_info=False, return_maxlocs=False,
                 do_co=False, redo_fa=True, missing_value=-9999.):
    r"""
    Upscales a river network by a factor (integer > 1), that has to be a
    divisible of the resolution of the flow direction. Direction is given by
    the cell with the largest flow accumulation. If multiple of these exist,
    then one is chosen randomly.

    Definition
    ----------
    upscale_fdir(sn, factor, print_info=False, return_maxlocs=False,
                 do_co=False, redo_fa=True)


    Input
    -----
    sn
        river_network object containing flow direction, flow accumulation and
        sinks
    factor
        integer indicating by which factor the flow direction should be
        upscaled

    Optional Input Parameters
    -------------------------
    print_info
        flag for printing additional information
    return_maxlocs
        flag for return locations of cell determining flow directions at given
        river_network object
    do_co
        flag for calculating channel order
    redo_fa
        flow recalculating flow accumulation at coarser river_network

    Options
    -------

    Output
    ------
    river_network object at coarser resolution with upscaled flow direction

    Restrictions
    ------------
    Given river_network object sn has to contain flow direction, accumulation
    and location of sink following the convention below.

    The origin of the flow direction field is assumed to be located in the
    upper left corner.
    Then flow directions are following this convention:

    Flow direction is assumed like this meaning that the ORIGIN IS THE UPPER
    LEFT CORNER
              64             y-axis
          32      128          |
      16      -1       1       |
           8       2          \|/
               4               V
         x-axis ------>

    Sinks are marked by -1.

    Examples
    --------
    >>> # Create some data
    >>> fd = np.ma.array([[  2,   1,   1,   2,   4,   4,   8,  8,  8],
    ...                   [  1,   2,   1,   1,   2,   4,   4,  4,  8],
    ...                   [128,   1, 128,   1,   1,   2,   4,  4,  4],
    ...                   [  1, 128,  64, 128, 128,   2,   4,  4,  8],
    ...                   [128,  64,  64,  64,   1,   2,   4,  4,  4],
    ...                   [ 64, 128,  64,  32,   1,   1,   2,  2,  4],
    ...                   [128,  64,  64,  64,   1,   1,   1,  1,  1],
    ...                   [128,  64, 128,  64,  32,   1, 128, 64, 64],
    ...                   [ 64, 128,  64,  64,  64, 128,  64, 64, 32]])
    >>> sinks = np.array([[6], [8]])
    >>> sn = river_network(fdir=fd, do_fa=True, do_co=False, sinks=sinks)
    >>> print(upscale_fdir(sn, 3).fdir)
    [[1.0 2.0 4.0]
     [64.0 16.0 4.0]
     [64.0 64.0 1.0]]


    History
    -------
    Written,  Stephan Thober, Feb 2016
    Modified, Matthias Cuntz, Nov 2016 - ported to Python 3
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    # consistency checks
    if not np.ma.isMaskedArray(sn.fa):
        raise ValueError('***ERROR: upscale_fdir requires flow accumulation as masked array in river_network')
    if any(np.array(sn.fdir.shape) % factor != 0):
        raise ValueError('***ERROR: factor: ' + str(factor) +
                         ' is not a divisible of flow direction shape')
    # create upscaled arrays
    new_shape = np.array(sn.fdir.shape) // factor
    new_fd = np.ma.masked_array(np.zeros(new_shape) + missing_value,
                                mask=np.ones(new_shape),
                                fill_value=missing_value)
    new_fa = np.ma.masked_array(np.zeros(new_shape) + missing_value,
                                mask=np.ones(new_shape),
                                fill_value=missing_value)

    # create maxlocs list
    maxlocs = []
    for ii in range(new_fd.shape[0]):
        for jj in range(new_fd.shape[1]):
            # extract part of map evaluated
            tmp_fa = sn.fa[ii * factor: (ii + 1) * factor,
                           jj * factor: (jj + 1) * factor]
            tmp_fd = sn.fdir[ii * factor: (ii + 1) * factor,
                             jj * factor: (jj + 1) * factor]
            if np.all(tmp_fa.mask):
                if print_info:
                    print('cell is masked ', ii, jj)
                # cell is masked
                new_fd[ii, jj] = -9999.
                new_fa[ii, jj] = -9999.
            else:
                # determine locations of maximum flow accumulations
                maxloc = np.ma.where(tmp_fa == np.ma.amax(tmp_fa))
                # calculate coarse scale flow direction
                coarse_fd = cal_fdir(maxloc, tmp_fd, factor)
                if maxloc[0].shape[0] > 1:
                    # if there is more than one outflow cell,
                    # check whether flow directions are different
                    if print_info:
                        print(coarse_fd, ' flow directions of same cells')
                    # evaluate when there are more than one cell
                    # if maximum flow directions are different
                    if np.any(np.diff(coarse_fd) > 0):
                        print('***Warning: multiple cells with same flow accumulation but different flow directions found, arbitrarily choose first one')
                # store flow direction and flow accumulation
                new_fd[ii, jj] = coarse_fd[0]
                new_fa[ii, jj] = tmp_fa[maxloc[0][0], maxloc[1][0]]
                if print_info:
                    print('maxloc = ', maxloc)
                    print('tmp_fd = ', tmp_fd[maxloc], coarse_fd[0])
                    print('tmp_fa = ', tmp_fa[maxloc])
                    print('----------------')
                    print('new_fd = ', new_fd[ii, jj])
                    print('new_fa = ', new_fa[ii, jj])
                    print('================')
                # add to store maximum locations
                maxlocs.append([maxloc[0] + ii * factor,
                                maxloc[1] + jj * factor])
    # upscale sinks
    upscale_sinks = tuple(np.array(sn.sinks)/int(factor))
    # return
    if return_maxlocs:
        if redo_fa:
            return maxlocs, river_network(fdir=new_fd, do_co=do_co, do_fa=True,
                                          sinks=upscale_sinks)
        else:
            return maxlocs, river_network(fdir=new_fd, do_co=do_co, fa=new_fa)
    else:
        if redo_fa:
            return river_network(fdir=new_fd, do_co=do_co, do_fa=True,
                                 sinks=upscale_sinks)
        else:
            return river_network(fdir=new_fd, do_co=do_co, fa=new_fa)


class river_network(object):
    def __init__(self, dem=None, fdir=None, co=None, do_co=False,
                 fa=None, do_fa=False, print_info=False, missing_value=-9999.,
                 sinks=None):
        r"""
        Initializes a river_network object describing the flow path of a river
        through the landscape.

        Definition
        ----------
        river_network(dem=None, fdir=None, co=None, do_co=False, fa=None,
                      do_fa=False, missing_value=-9999., sinks=None):


        Input
        -----
        dem
            digital elevation model (dem), basis for calculating flow
            direction fdir
        fdir
            flow direction following convention below.

        Optional Input Parameters
        -------------------------
        co
            channel order following Strahler 1952
        do_co
            flag for calculating channel order
        fa
            flow accumulation
        do_fa
            flag for calculating flow accumulation
        print_info
            flag for printing additional information
        missing_value
            default: -9999.
        sinks
            location of sinks as two arrays (first/second for y/x coordinate)

        Restrictions
        ------------
        Either dem or fdir has to be provided, both cannot be omitted.

        The origin of the flow direction field is assumed to be located in the
        upper left corner.
        Then flow directions are following this convention:

        Flow direction is assumed like this meaning that the ORIGIN IS THE
        UPPER LEFT CORNER
                  64             y-axis
              32      128          |
          16      -1       1       |
               8       2          \|/
                   4               V
             x-axis ------>

        Sinks are marked by -1.

        Examples
        --------
        >>> # Create some data
        >>> fd = np.ma.array([[  2,   1,   1,   2,   4,   4,   8,  8,  8],
        ...                   [  1,   2,   1,   1,   2,   4,   4,  4,  8],
        ...                   [128,   1, 128,   1,   1,   2,   4,  4,  4],
        ...                   [  1, 128,  64, 128, 128,   2,   4,  4,  8],
        ...                   [128,  64,  64,  64,   1,   2,   4,  4,  4],
        ...                   [ 64, 128,  64,  32,   1,   1,   2,  2,  4],
        ...                   [128,  64,  64,  64,   1,   1,   1,  1,  1],
        ...                   [128,  64, 128,  64,  32,   1, 128, 64, 64],
        ...                   [ 64, 128,  64,  64,  64, 128,  64, 64, 32]])
        >>> sinks = np.array([[6], [8]])
        >>> print(river_network(fdir=fd, do_fa=True, do_co=False, sinks=sinks).fa)
        [[1.0 1.0 2.0 3.0 1.0 1.0 1.0 1.0 1.0]
         [1.0 4.0 1.0 32.0 37.0 3.0 2.0 2.0 1.0]
         [1.0 1.0 30.0 1.0 4.0 46.0 3.0 4.0 1.0]
         [1.0 5.0 19.0 2.0 1.0 1.0 50.0 5.0 2.0]
         [2.0 1.0 18.0 1.0 1.0 2.0 52.0 8.0 1.0]
         [1.0 6.0 2.0 9.0 1.0 2.0 57.0 9.0 2.0]
         [1.0 4.0 1.0 8.0 1.0 2.0 3.0 68.0 81.0]
         [2.0 1.0 3.0 2.0 2.0 1.0 4.0 3.0 1.0]
         [1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0]]
        >>> print(river_network(fdir=fd, do_fa=False, do_co=True, sinks=sinks).co)
        [[1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0]
         [1.0 2.0 1.0 3.0 3.0 2.0 1.0 1.0 1.0]
         [1.0 1.0 3.0 1.0 2.0 3.0 1.0 2.0 1.0]
         [1.0 2.0 3.0 1.0 1.0 1.0 3.0 2.0 1.0]
         [1.0 1.0 3.0 1.0 1.0 1.0 3.0 2.0 1.0]
         [1.0 2.0 1.0 2.0 1.0 1.0 3.0 2.0 1.0]
         [1.0 2.0 1.0 2.0 1.0 1.0 1.0 3.0 3.0]
         [1.0 1.0 2.0 1.0 1.0 1.0 2.0 2.0 1.0]
         [1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0 1.0]]


        History
        -------
        Written,  Stephan Thober, Dec 2015
        Modified, Matthias Cuntz, Sep 2021 - code refactoring
        """
        # initialize all arrays
        self.dem = None    # digital elevation model
        self.fdir = None   # flow direction
        self.sinks = None  # sinks
        self.co = None  # channel order
        self.fa = None  # flow accumulation
        # consistency check
        if fdir is None  and dem is None:
            raise ValueError('***ERROR: specify either dem or fdir to create a river_network object')
        if fdir is None:
            if print_info:
                print('calculating flow direction')
            # create flow direction if necessary
            self.dem = dem
            self.fdir = self.flow_direction()
            if print_info:
                print('calculating flow direction... ok')
        else:
            self.fdir = fdir
        # assure that fdir is masked array
        if not np.ma.isMaskedArray(self.fdir):
            self.fdir = np.ma.array(self.fdir)
            self.fdir.mask = False
        # assign flow accumulation
        if fa is not None:
            self.fa = fa
        # assign channel order
        if co is not None:
            self.co = co
        # assign sinks
        if sinks is None and fa is None:
            raise ValueError('***ERROR: for initializing river network either the location of the sinks or flow accumulation has to be given')
        elif sinks is None and fa is not None:
            self.sinks = self._get_sinks()
        elif sinks is not None and fa is None:
            self.sinks = sinks
        # get channel order and flow accumulation
        if do_co and do_fa:
            self.co = np.ma.masked_array(np.zeros(self.fdir.shape) +
                                         missing_value,
                                         mask=np.ones(self.fdir.shape),
                                         fill_value=missing_value)
            self.fa = np.ma.masked_array(np.zeros(self.fdir.shape) +
                                         missing_value,
                                         mask=np.ones(self.fdir.shape),
                                         fill_value=missing_value)
            for ii in range(self.sinks[0].shape[0]):
                self.co, self.fa = self.network_properties(
                    self.fdir, self.sinks[0][ii], self.sinks[1][ii],
                    do_co=do_co, co=self.co, do_fa=do_fa, fa=self.fa,
                    missing_value=missing_value, print_info=print_info)
        elif do_co and not do_fa:
            self.co = np.ma.masked_array(
                np.zeros(self.fdir.shape) + missing_value,
                mask=np.ones(self.fdir.shape), fill_value=missing_value)
            for ii in range(self.sinks[0].shape[0]):
                self.co = self.network_properties(
                    self.fdir, self.sinks[0][ii], self.sinks[1][ii],
                    do_co=do_co, co=self.co, do_fa=do_fa,
                    missing_value=missing_value, print_info=print_info)
        elif not do_co and do_fa:
            self.fa = np.ma.masked_array(
                np.zeros(self.fdir.shape) + missing_value,
                mask=np.ones(self.fdir.shape), fill_value=missing_value)
            for ii in range(self.sinks[0].shape[0]):
                self.fa = self.network_properties(
                    self.fdir, self.sinks[0][ii], self.sinks[1][ii],
                    do_co=do_co, do_fa=do_fa, fa=self.fa,
                    missing_value=missing_value, print_info=print_info)

    def flow_direction(self, print_info=False):
        r"""
        Calculates flow direction from a DEM.

        Definition
        ----------
        def flow_direction(self, print_info=False)


        Input
        -----
        self         self - river_network object containing a dem array

        Optional Input Parameters
        -------------------------
        print_info - flag for printing additional information

        Options
        -------

        Output
        ------
        fd
            array containing flow direction with the following convention

        Restrictions
        ------------
        The origin of the flow direction field is assumed to be located in the
        upper left corner.
        Then flow directions are following this convention:

        flow direction is assumed like this meaning that the ORIGIN IS THE
        UPPER LEFT CORNER
                  64             y-axis
              32      128          |
          16      -1       1       |
               8       2          \|/
                   4               V
             x-axis ------>

        Sinks are marked by -1.

        Examples
        --------
        >>> # Create some data
        >>> dem = np.ma.array([[ 30,   2,   1],
        ...                    [  5,  10,  25],
        ...                    [ 15,  23,  24]])
        >>> dem.mask = np.zeros(dem.shape, dtype='bool')
        >>> print(river_network(dem=dem, do_fa=False, do_co=False, sinks=np.array([[0], [2]])).fdir)
        [[1.0 1.0 -1.0]
         [128.0 128.0 64.0]
         [64.0 32.0 32.0]]


        History
        -------
        Written,  Stephan Thober & David Schaefer, Dec 2015
        Modified, Matthias Cuntz, Sep 2021 - code refactoring
        """
        # global variable used: correct_direction
        fd = np.zeros(self.dem.shape)
        #
        for ii in range(self.dem.shape[0]):
            for jj in range(self.dem.shape[1]):
                if self.dem.mask[ii, jj]:
                    continue
                if print_info:
                    print('processing cell: ', ii, jj)
                # get mask of neighbors and y and x locations
                neighbors, yy, xx = self._get_neighbors(self.dem, ii, jj)
                # get position of cell with steepest gradient
                pos_min = np.ma.argmin(self.dem[yy, xx] - self.dem[ii, jj])
                fd[ii, jj] = local_flow_direction[neighbors][pos_min]
        return fd

    def network_properties(self, fd, yy, xx, print_info=False, do_co=True,
                           co=None, do_fa=True, fa=None,
                           missing_value=-9999.):
        r"""
        Calculates channel order number and flow accumulation starting from one
        sink in a flow direction map

        channel order is calculated following Strahler 1952. It is ONE for
        headwater. If channels join, the channel order of the resulting stream
        is the highest one of the inflowing streams, if two or more than two
        inflowing streams have the highest channel order, the channel order of
        the resulting stream is one higher than the highest channel order of
        the inflowing streams.

        Definition
        ----------
        def network_properties(self, fd, yy, xx, print_info=False, do_co=True,
        co=None, do_fa=True, fa=None, missing_value=-9999.):


        Input
        -----
        self          self - river_network object
        fd            flow direction field, basically river_network.fd
        yy            row coordinate of sink
        xx            column coordinate of sink


        Optional Input Parameters
        -------------------------
        print_info    write additional info on std_out
        do_co         calculate channel order
        co            given channel order field
        do_fa         calculate flow accumulation
        fa            given flow accumulation
        missing_value floating point value for masking

        Options
        -------
        print_info   True:  write additional information
                     False: do not write additional information (default)
        do_co        True:  calculate channel order (default)
                     False: do not channel order
        co           None:  no channel order field specified,
                            will be created (default)
        do_fa        True:  calculate flow accumulation (default)
                     False: do not flow accumulation
        fa           None:  no flow accumulation field specified,
                            will be created (default)

        Output
        ------
        Depending on options:
            co, fa if do_co=True and do_fa=True
            co if do_co=True and not do_fa=True
            fa if not do_co=True and do_fa=True


        Restrictions
        ------------
        The origin of the flow direction field is assumed to be located in the
        upper left corner.
        Then flow directions are following this convention:

        flow direction is assumed like this meaning that the ORIGIN IS THE
        UPPER LEFT CORNER
                  64             y-axis
              32      128          |
          16      -1       1       |
               8       2          \|/
                   4               V
             x-axis ------>

        Sinks are marked by -1.


        History
        -------
        Written,  Stephan Thober & David Sschaefer, Dec 2015
        Modified, Matthias Cuntz, Nov 2016 - ported to Python 3
                  Matthias Cuntz, Oct 2018 - int indices in fa[]
                  Matthias Cuntz, Sep 2021 - code refactoring
        """
        if co is None and do_co:
            co = np.ma.masked_array(
                np.zeros(fd.shape) + missing_value, mask=np.ones(fd.shape),
                fill_value=missing_value)
        if fa is None and do_fa:
            fa = np.ma.masked_array(
                np.zeros(fd.shape) + missing_value, mask=np.ones(fd.shape),
                fill_value=missing_value)
        # flow direction stack to emulate recursion
        if not do_co and not do_fa:
            raise ValueError('***ERROR: neither fa nor co calculated')
        fd_stack = [[yy, xx]]  # start at initial sink
        while fd_stack:
            if print_info:
                print('current flow accumulation stack: ', fd_stack)
            upstream = self._get_upstream(fd, fd_stack[-1])
            if print_info:
                print('upstream locations: ', upstream)
            if do_co:
                # use co for identifying upstream cells
                ext = [l for l in upstream
                       if co.data[int(l[0]), int(l[1])] == missing_value]
            else:
                # use fa for identifying upstream cells
                ext = [l for l in upstream
                       if fa.data[int(l[0]), int(l[1])] == missing_value]
            if ext:
                fd_stack.extend(ext)
                continue
            # all upstream cells are available
            # note that headwaters dont have an upstream cell
            cell = fd_stack.pop()  # save cell
            if do_co:
                co_upstream = [co[loc[0], loc[1]] for loc in upstream]
                co_max = np.amax(co_upstream + [1])
                # if two streams of equal co merge, increment channel order
                if len(np.where(co_upstream == co_max)[0]) > 1:
                    co_max += 1
                co.data[int(cell[0]), int(cell[1])] = co_max
                co.mask[int(cell[0]), int(cell[1])] = False
                if print_info:
                    print('co (channel order) of upstream: ', co_upstream)
            if do_fa:
                fa_upstream = [ fa[int(loc[0]), int(loc[1])]
                                for loc in upstream ]
                fa.data[int(cell[0]), int(cell[1])] = np.sum(fa_upstream) + 1
                fa.mask[int(cell[0]), int(cell[1])] = False
                if print_info:
                    print('sum of upstream: ', np.sum(fa_upstream))
        if do_co and do_fa:
            return co, fa
        elif do_co and not do_fa:
            return co
        elif not do_co and do_fa:
            return fa

    def _get_sinks(self):
        # set sinks to maximum flow accumulation
        return np.ma.where(self.fa == np.amax(self.fa))

    def _get_neighbors(self, arr, yy_loc, xx_loc):
        # global variables used: yy_offset, xx_offset
        #
        yy_ind = yy_offset + yy_loc
        xx_ind = xx_offset + xx_loc
        # create mask for valid neighbors
        neighbors = ((yy_ind >= 0) &
                     (yy_ind < arr.shape[0]) &
                     (xx_ind >= 0) &
                     (xx_ind < arr.shape[1]))
        return neighbors, yy_ind[neighbors], xx_ind[neighbors]

    def _get_upstream(self, fd, loc):
        # global variable used: inflow_direction
        #
        # get mask of neighbors and y and x locations
        neighbors, yy, xx = self._get_neighbors(fd, loc[0], loc[1])
        # mask inflowing cells
        upstream_mask = (fd.data[yy.astype(int), xx.astype(int)] ==
                         inflow_direction[neighbors])
        yy_upstream = yy[upstream_mask]
        xx_upstream = xx[upstream_mask]
        return [ [yy_upstream[ii], xx_upstream[ii]]
                 for ii in range(np.sum(upstream_mask)) ]


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
from functools import partial
import subprocess
import numpy as np
from pyjams.const import huge
# ToDo:
#   Handling constraints
#   write tmp/population files (as in SCE of Fortran)
#   write out also in logfile if not None (use jams.tee as in joptimise)
#   crossover with quadratic function (QIPSO) ?


def _ext_obj_wrapper(func, lb, ub, mask,
                     parameterfile, parameterwriter, objectivefile, objectivereader, shell, debug, rank,
                     params):
    '''
        Wrapper function for external program to be optimised
        to be used with partial:
            obj = partial(_ext_obj_wrapper, func, lb, ub, mask,
                          parameterfile, parameterwriter, objectivefile, objectivereader, shell, debug, rank)
        This allows then calling obj with only the argument params:
            fx = obj(params)


        Definition
        ----------
        def _ext_obj_wrapper(func, lb, ub, mask,
                             parameterfile, parameterwriter, objectivefile, objectivereader, shell, debug, rank,
                             params):


        Input
        -----
        func              function to minimise (python function or string for external executable)
        lb                (npars) lower bounds of parameters
        ub                (npars) upper bounds of parameters
        mask              (npars) mask to include (1) or exclude (0) parameter from optimisation
        parameterfile     Parameter file for executable; must be given if func is name of executable
        parameterwriter   Python function for writing parameter file if func is name of executable
        objectivefile     File with objective value from executable; must be given if func is name of executable
        objectivereader   Python function for reading objective value if func is name of executable
        shell             If True, the specified command will be executed through the shell.
        debug             If True, model output is displayed for executable.
        rank              Process rank will be given as last argument to external program.
        params            (npars) parameter set


        Output
        ------
        Function value


        History
        -------
        Written,  MC, Nov 2016
    '''
    if isinstance(func, (str,list)):
        parameterwriter(parameterfile, params, lb, ub, mask)
        if rank is None:
            func1 = func
        else:
            if isinstance(func, str):
                func1 = [func, str(rank)]
            else:
                func1 = func+[str(rank)]
        if debug:
            err = subprocess.call(func1, shell=shell)
        else:
            err = subprocess.check_output(func1, shell=shell)
        obj = objectivereader(objectivefile)
        return obj
    else:
        return func(params)


# Function wrappers for objective and constraints
# used with functools.partial
def _obj_wrapper(func, arg, kwarg, x):
    '''
        Wrapper function for function to be optimised
        to be used with partial:
            obj = partial(_obj_wrapper, func, arg, kwarg)
        This allows then calling obj with only the argument x:
            fx = obj(x)
        which translates to:
            fx = func(x, *arg, **kwarg)
    '''
    return func(x, *arg, **kwarg)


def _is_feasible_wrapper(func, x):
    '''
        Checking that all constraints are met.

        func is one of the 'partialised' constraint functions
            _cons_none_wrapper, _cons_ieqcons_wrapper, and _cons_f_ieqcons_wrapper.

        This function is partialised:
            is_feasible = partial(_is_feasible_wrapper, cons)
        which is unnecessary for now. It could still simply be:
            is_feasible = cons
    '''
    return np.all(func(x)>=0.)


def _cons_none_wrapper(x):
    '''
        Dummy constraints function if no constraints.
        Returns 0.
    '''
    return np.array([0])


def _cons_ieqcons_wrapper(ieqcons, arg, kwarg, x):
    '''
        Wrapper function for constraints to be used with partial in the case
        that ieqcons is given, i.e. a list of constraint functions, returning
        >=0. if constraints are met.
    '''
    return np.array([y(x, *arg, **kwarg) for y in ieqcons])


def _cons_f_ieqcons_wrapper(f_ieqcons, arg, kwarg, x):
    '''
        Wrapper function for constraints to be used with partial in the case
        that f_ieqcons is given, i.e. a single function returning an 1-D array
        values >=0. where constraints are met.
    '''
    return np.array(f_ieqcons(x, *arg, **kwarg))


def range012range(x, xmin, xmax):
    '''
        Convert range between [0,1] to range [xmin,xmax]
    '''
    return xmin + x*(xmax-xmin)


def range2range01(x, xmin, xmax):
    '''
        Convert range between [xmin,xmax] to range [0,1]
    '''
    return (x-xmin)/(xmax-xmin)

    
def diversity(x, lb, ub, mask):
    '''
        Diversity in swarm positions.
        From [Pant et al., 2009] http://www.iaeng.org/IJCS/issues_v36/issue_2/IJCS_36_2_02.pdf
        changed to RMSE.

        RMSE is generally about 1/2 or 1/3 of normalized geometric range (SCE).
    '''
    # all dimensions [0,1]
    x01 = np.where(mask, range2range01(x,lb,ub), x)

    S = float(x01.shape[0])
    D = float(x01.shape[1])

    # average particle position
    pmean = np.sum(x01, axis=0)/S

    # # Eq. 5 of Pant et al. (2009)
    # div = np.sum(np.sqrt(np.sum((x01-pmean)**2,axis=1)),axis=0) / S

    # average RMSE
    div = np.sum(np.sqrt(np.sum((x01-pmean)**2,axis=1)/D),axis=0) / S

    # # Normalized geometric range of SCE
    # imax = x01.max(axis=0)
    # imin = x01.min(axis=0)
    # imaxmin = np.ma.array(imax-imin, mask=(imax==imin))
    # div = np.ma.exp(np.ma.mean(np.ma.log(imaxmin)))

    return div


def get_neighbor_indeces(n, S, topology, kl=1):
    '''
        Get the indices of the neighbors for the current particle given the topology.

        Input
        -----
        n           integer
                    Particle index.
        S           integer
                    Swarm size.
        topology    string
                    Neighborhood topologies. These are rather social than geographical topologies.
                    All neighborhoods comprise the current particle as well.
                    [Kennedy & Mendes, 2002] doi: 10.1109/CEC.2002.1004493
                    'gbest'    Neighborhood is entire swarm.
                    'lbest'    Partciles arranged in a ring, in which each particle communicates with
                               kl particles on each side, i.e. particle i has the neighborhood
                               i-kl, i-kl+1, ..., i, i+1, ..., i+kl-1, i+kl
                               [Mohais et al., 2005] doi: 10.1007/11589990_80
                    'neumann'  Neighborhood of a point including all points at a Hamming distance of 1.
                               Particles are arranges in a lattice, where each particle interacts with
                               its immediate 4 neighbors to the N, S, E, and W.
                               [Kennedy and Mendes, 2006] doi: 10.1109/TSMCC.2006.875410
                               The von Neumann neighborhood is configured into r rows and c columns,
                               where r is the highest integer less than or equal to sqrt(n) that evenly
                               divides n and c = n / r
                               [Mohais et al., 2005] doi: 10.1007/11589990_80
                    'ring'     'lbest' with kl=1


        Optional Input
        --------------
        kl          integer
                    Neighborhood distance in topology 'lbest'.
                    (Default: 1 = ring)
    '''
    if topology.lower() == 'ring': kl=1

    if topology.lower() == 'gbest':
        ii = range(S)
    elif (topology.lower() == 'lbest') or (topology.lower() == 'ring'):
        ii = [n]                  # self
        for kk in range(1,kl+1):
            ii.append((n-kk) % S) # left
            ii.append((n+kk) % S) # right
    elif topology.lower() == 'neumann':
        rows = int(np.floor(np.sqrt(S)))
        while (S % rows) != 0: rows -= 1
        cols = S // rows
        left  = lambda x, c: (x - 1) % c + x//c * c
        right = lambda x, c: (x + 1) % c + x//c * c
        above = (n - rows) % S
        below = (n + rows) % S
        ii = [left(above, cols),  above, right(above, cols),
              left(n, cols),      n,     right(n, cols),
              left(below, cols),  below, right(below, cols)]

    return ii


def get_best_neighbor(p, fp, topology, kl=1):
    '''
        Get the best neighbor for a given topology

        Input
        -----
        p           ND-array
                    The best known position of each particle.
        fp          1D-array
                    The objective values at each position in p.
        topology    string
                    Neighborhood topologies. These are rather social than geographical topologies.
                    All neighborhoods comprise the current particle as well.
                    [Kennedy & Mendes, 2002] doi: 10.1109/CEC.2002.1004493
                    'gbest'    Neighborhood is entire swarm.
                    'lbest'    Partciles arranged in a ring, in which each particle communicates with
                               kl particles on each side, i.e. particle i has the neighborhood
                               i-kl, i-kl+1, ..., i, i+1, ..., i+kl-1, i+kl
                               [Mohais et al., 2005] doi: 10.1007/11589990_80
                    'neumann'  Neighborhood of a point including all points at a Hamming distance of 1.
                               Particles are arranges in a lattice, where each particle interacts with
                               its immediate 4 neighbors to the N, S, E, and W.
                               [Kennedy and Mendes, 2006] doi: 10.1109/TSMCC.2006.875410
                               The von Neumann neighborhood is configured into r rows and c columns,
                               where r is the highest integer less than or equal to sqrt(n) that evenly
                               divides n and c = n / r
                               [Mohais et al., 2005] doi: 10.1007/11589990_80
                    'ring'     'lbest' with kl=1


        Optional Input
        --------------
        kl          integer
                    Neighborhood distance in topology 'lbest'.
                    (Default: 1 = ring)
    '''
    if topology.lower() == 'ring': kl=1
    S = p.shape[0]
    D = p.shape[1]

    g  = np.ones((S,D))*np.inf
    fg = np.ones(S)*np.inf
    if topology.lower() == 'gbest':
        i_min = np.argmin(fp)
        ig  = p[i_min,:] # overall best
        ifg = fp[i_min]
        for ss in range(S):
            g[ss,:] = ig
            fg[ss]  = ifg
    elif (topology.lower() == 'lbest') or (topology.lower() == 'ring') or (topology.lower() == 'neumann'):
        g  = np.ones((S,D))*np.inf
        fg = np.ones(S)*np.inf
        for ss in range(S):
            ii = get_neighbor_indeces(ss, S, topology, kl=kl)
            pp  = p[ii,:]
            fpp = fp[ii]
            i_min = np.argmin(fpp)
            g[ss,:] = pp[i_min,:].copy()
            fg[ss]  = fpp[i_min]

    return [g, fg]


def rwde(func, feasible, x, fx, xmin, xmax, lb, ub, x0, mask, maxit, nlocal=5):
    '''
        Random Walk with Direction Exploitation
        [Petalas et al. 2007] doi: 10.1007/s10479-007-0224-y
        as implemented in
        [Wang et al. 2012] doi: 10.1016/j.ins.2012.02.016


        Input
        -----
        func        objective function wrapper
                    function to min-/maximise
        feasible    constraint function wrapper
        x           1D-array
                    Position of one particle in parameter space.
        fx          scalar
                    The objective value at position x.
        xmin        1D-array
                    Lower bounds of particles search positions (parameters).
        xmax        1D-array
                    Upper bounds of particles search positions (parameters).
        lb          1D-array
                    The lower bounds of the particle positions (parameters).
        ub          1D-array
                    The upper bounds of the particle positions (parameters).
        x0          1D-array
                    Will be taken at dimensions with mask==False.
        mask        1D-array
                    include (1,True) or exclude (0,False) parameters in optimisation.


        Optional Input
        --------------
        nlocal      numer of local steps (Default: 5)
    '''
    # constants
    D = x.shape[0]
    lam_init = 1. # inital stepsize
    # local search
    large = fx
    t     = 0
    lam   = lam_init
    ngood = 0
    for t in range(nlocal):
        z    = np.random.uniform(size=D)
        xnew = x + lam*z*(xmax - xmin)
        xnew = np.where(mask, xnew, x0) # mask
        xnew = np.clip(xnew, lb, ub)    # limit
        fs   = feasible(xnew)
        if fs:
            fxnew = func(xnew)
            if maxit: fxnew *= -1.
            # NaN/Inf
            large = 1.1*large if large>0. else 0.9*large
            fxnew = fxnew if np.isfinite(fxnew) else large
            if fxnew < fx:
                x   = xnew
                fx  = fxnew
                lam = lam_init
                ngood += 1
            elif fxnew > fx:
                lam *= 0.5
            else:
                continue

    return x, fx, ngood


def cbls(func, feasible, x, fx, p, xmin, xmax, lb, ub, x0, mask, inertia, phip, maxit, strategy, nlocal=5):
    '''
        Cognition-Based Local Search
        [Wang et al. 2012] doi: 10.1016/j.ins.2012.02.016


        Input
        -----
        func        objective function wrapper
                    function to mini-/maximise
        feasible    constraint function wrapper
        x           1D-array
                    Position of one particle in parameter space.
        fx          1D-array
                    The objective value at position x.
        p           1D-array
                    Particle x' best position.
        xmin        1D-array
                    Lower bounds of particles search positions (parameters).
        xmax        1D-array
                    Upper bounds of particles search positions (parameters).
        lb          1D-array
                    The lower bounds of the particle positions (parameters).
        ub          1D-array
                    The upper bounds of the particle positions (parameters).
        x0          1D-array
                    Will be taken at dimensions with mask==False.
        mask        1D-array
                    include (1,True) or exclude (0,False) parameters in optimisation.
        inertia     scalar
                    Particle velocity scaling factor.
                    Default depends on algorithm (strategy).
        phip        scalar
                    Scaling factor to search away from the particle's best known position.
                    Default depends on algorithm (strategy).
        strategy    string
                    PSO variants.
                    'original':   Textbook particle swarm algorithm with inertia weight
                                  x = current position
                                  p = particles best position
                                  g = neighborhood best position
                                  rg, rp = np.random.uniform(size=(S,D))
                                  inertia=0.5, phip=2., phig=2.
                                  v = inertia*v + rp*phip*(p-x) + rg*phig*(g-x)
                                  x = x + v
                    'inertia':    Same as 'original' but with inertia weight decreasing from 0.9 to 0.4
                                  over time, i.e. over iterations (it)
                                  imax = 0.9
                                  imin = 0.4
                                  inertia = imax - float(it)/float(maxn-1) * (imax-imin)
                    'canonical':  Clerk & Kennedy (2000) with fixed, clamped constriction factor
                                  From PaGMO (esa.github.io/pagmo):
                                  Clerc's analysis of the iterative system led him to propose a strategy for the
                                  placement of "constriction coefficients" on the terms of the formulas; these
                                  coefficients controlled the convergence of the particle and allowed an elegant and
                                  well-explained method for preventing explosion, ensuring convergence, and
                                  eliminating the arbitrary vmax parameter. The analysis also takes the guesswork
                                  out of setting the values of phi_1 and phi_2.
                                  "This is the canonical particle swarm algorithm of today."
                                  [Poli et al., 2007] doi: 10.1007/s11721-007-0002-0
                                  [Clerc & Kennedy, 2002] doi: 10.1109/4235.985692
                                  From F van den Bergh, Diss 2001:
                                  The problem [with the constriction factor], according to Eberhart and Shi,
                                  is that the particles stray too far from the desired region of search space.
                                  To mitigate this effect they decided to apply clamping to the constriction factor
                                  implementation as well, setting the vmax parameter equal to xmax, the size of the
                                  search space. This led to improved performance for almost all the functions they
                                  used during testing - both in terms of the rate of convergence and the ability of
                                  the algorithm to reach the error threshold.
                                  [Eberhardt RC & Shi Y, 2000] Comparing inertia weights and constriction factors
                                  in particle swarm optimization. In: Proceedings of the Congress on Evolutionary
                                  Computation (CEC2000), 84-88.
                                  [van den Bergh F, 2000] An Analysis of Particle Swarm Optimizers, PhD thesis,
                                  University of Pretoria, South Africa
                                  inertia=0.7289, phip=2.05, phig=2.05
                                  v = inertia * (v + phip*rp*(p - x) + phig*rg*(g - x))
                                  v = clip(v, lb, ub)
                                  x = x + v


        Optional Input
        --------------
        nlocal      numer of local steps (Default: 5)
    '''
    # Strategy keyword
    ptypes = ['original', 'inertia', 'canonical']
    assert strategy.lower() in ptypes, 'Cognition-Based Local Search not available for strategy {:}'.format(strategy)
    # constants
    D = x.shape[0] # # parameters
    
    large = fx
    fxnew = fx
    xnew  = x[:]
    v     = xmin + np.random.uniform(size=D)*(xmax-xmin)
    ngood = 0
    for t in range(nlocal):
        # update velocity with only cognitive and without social component
        rp = np.random.uniform(size=D)
        if (strategy.lower() == 'original') or (strategy.lower() == 'inertia'):
            vnew = inertia*v + phip*rp*(p - xnew)
        elif strategy.lower() == 'canonical':
            vnew = inertia * (v + phip*rp*(p - xnew))
        # new x
        xnewtest = xnew + vnew                  # tentative update of x
        xnewtest = np.where(mask, xnewtest, x0) # mask
        xnewtest = np.clip(xnewtest, lb, ub)    # limit
        # new fx
        fs = feasible(xnewtest)
        if fs:
            fxnewtest = func(xnewtest)
            if maxit: fxnewtest *= -1.
            # NaN/Inf
            large = 1.1*large if large>0. else 0.9*large
            fxnewtest = fxnewtest if np.isfinite(fxnewtest) else large
            if fxnewtest < fx:
                xnew  = xnewtest
                fxnew = fxnewtest
                ngood += 1

    return xnew, fxnew, ngood


# Particle Swarm Optimisation
def pso(func, x0, lb, ub,
        mask=None,
        ieqcons=None, f_ieqcons=None,
        arg=(), kwarg={},
        swarmsize=None, inertia=None, phip=None, phig=None, maxn=250,
        minstep=1e-8, minobj=1e-8, maxit=False,
        init='lhs', strategy='canonical', topology='gbest', kl=1,
        memetic='no', nmemetic=1, nlocal=5, rrwde=0.01, pls=0.2,
        includex0=False, seed=None,
        processes=1,
        verbose=0, pout=False, cout=False,
        restart=False, restartfile1='pso.restart.npz', restartfile2='pso.restart.txt',
        parameterfile=None, parameterwriter=None,
        objectivefile=None, objectivereader=None,
        shell=False, debug=False):
    """
        Particle Swarm Optimization (PSO)


        Definition
        ----------
        def pso(func, x0, lb, ub,
                mask=None,
                ieqcons=None, f_ieqcons=None,
                arg=(), kwarg={},
                swarmsize=None, inertia=None, phip=None, phig=None, maxn=250,
                minstep=1e-8, minobj=1e-8, maxit=False,
                init='lhs', strategy='canonical', topology='gbest', kl=1,
                includex0=False, seed=None,
                processes=1,
                verbose=0, pout=False, cout=False,
                restart=False, restartfile1='pso.restart.npz', restartfile2='pso.restart.txt',
                parameterfile=None, parameterwriter=None,
                objectivefile=None, objectivereader=None,
                shell=False, debug=False):


        Input
        -----
        func        python function or string for external executable
                    The function to be minimized.
        x0          1D-array
                    Will be taken at dimensions with mask==False.
        lb          1D-array
                    The lower bounds of the particle positions (parameters).
        ub          1D-array
                    The upper bounds of the particle positions (parameters).


        Optional Input
        --------------
        mask        1D-array
                    include (1,True) or exclude (0,False) parameters in optimisation.
                    (Default: include all dimensions)
        ieqcons     list
                    A list of functions of length n such that ieqcons[j](x,*arg) >= 0.0 in
                    a successfully optimized problem.
                    (Default: None)
        f_ieqcons   function
                    Returns a 1-D array in which each element must be greater or equal
                    to 0.0 in a successfully optimized problem. If f_ieqcons is specified,
                    ieqcons is ignored.
                    (Default: None)
        arg         tuple
                    Additional arguments passed to objective and constraint functions (only for Python functions).
                    (Default: empty tuple)
        kwarg       dict
                    Additional keyword arguments passed to objective and constraint functions (only for Python functions).
                    (Default: empty dict)
        swarmsize   int
                    The number of particles in the swarm.
                    (Default: max(min(3*len(lb),40),10))
        inertia     scalar
                    Particle velocity scaling factor.
                    Default depends on algorithm (strategy).
        phip        scalar
                    Scaling factor to search away from the particle's best known position.
                    Default depends on algorithm (strategy).
        phig        scalar
                    Scaling factor to search away from the swarm's (neighbor's) best known position.
                    Default depends on algorithm (strategy).
        maxn        int
                    The maximum number of iterations for the swarm to search.
                    (Default: 250)
        minstep     scalar
                    The minimum stepsize of swarm's best position before the search terminates.
                    (Default: 1e-8)
        minobj      scalar
                    Objective function defining convergence. Attention at maxit=True.
                    (Default: 1e-8)
        maxit       boolean
                    Minimise or maximise func.
                    (Default: False)
                    False: minimise objective function down to minobj
                    True:  maximise objective function, i.e. minimise -objective function down to minobj
        init        string
                    How to sample the initial swarm positions and velocities.
                    (Default: 'lhs')
                    'random': random sampling from uniform distribution
                    'lhs':    latin hypercube sampling from uniform distributions
                    'sobol':  quasirandom Sobol sequence (only up to 40 dimensions)
        strategy    string
                    PSO variants.
                    (Default: 'canonical')
                    'original':   Textbook particle swarm algorithm with inertia weight
                                  x = current position
                                  p = particles best position
                                  g = neighborhood best position
                                  rg, rp = np.random.uniform(size=(S,D))
                                  inertia=0.5, phip=2., phig=2.
                                  v = inertia*v + rp*phip*(p-x) + rg*phig*(g-x)
                                  x = x + v
                    'inertia':    Same as 'original' but with inertia weight decreasing from 0.9 to 0.4
                                  over time, i.e. over iterations (it)
                                  imax = 0.9
                                  imin = 0.4
                                  inertia = imax - float(it)/float(maxn-1) * (imax-imin)
                    'canonical':  Clerk & Kennedy (2000) with fixed, clamped constriction factor
                                  From PaGMO (esa.github.io/pagmo):
                                  Clerc's analysis of the iterative system led him to propose a strategy for the
                                  placement of "constriction coefficients" on the terms of the formulas; these
                                  coefficients controlled the convergence of the particle and allowed an elegant and
                                  well-explained method for preventing explosion, ensuring convergence, and
                                  eliminating the arbitrary vmax parameter. The analysis also takes the guesswork
                                  out of setting the values of phi_1 and phi_2.
                                  "This is the canonical particle swarm algorithm of today."
                                  [Poli et al., 2007] doi: 10.1007/s11721-007-0002-0
                                  [Clerc & Kennedy, 2002] doi: 10.1109/4235.985692
                                  From F van den Bergh, Diss 2001:
                                  The problem [with the constriction factor], according to Eberhart and Shi,
                                  is that the particles stray too far from the desired region of search space.
                                  To mitigate this effect they decided to apply clamping to the constriction factor
                                  implementation as well, setting the vmax parameter equal to xmax, the size of the
                                  search space. This led to improved performance for almost all the functions they
                                  used during testing - both in terms of the rate of convergence and the ability of
                                  the algorithm to reach the error threshold.
                                  [Eberhardt RC & Shi Y, 2000] Comparing inertia weights and constriction factors
                                  in particle swarm optimization. In: Proceedings of the Congress on Evolutionary
                                  Computation (CEC2000), 84-88.
                                  [van den Bergh F, 2000] An Analysis of Particle Swarm Optimizers, PhD thesis,
                                  University of Pretoria, South Africa
                                  inertia=0.7289, phip=2.05, phig=2.05
                                  v = inertia * (v + phip*rp*(p - x) + phig*rg*(g - x))
                                  v = clip(v, lb, ub)
                                  x = x + v
                    'fips':       Fully Informed Particle Swarm
                                  From PaGMO (esa.github.io/pagmo):
                                  Whereas in the traditional algorithm each particle is affected by its own
                                  previous performance and the single best success found in its neighborhood, in
                                  Mendes' fully informed particle swarm (FIPS), the particle is affected by all its
                                  neighbors, sometimes with no influence from its own previous success.
                                  "With good parameters, FIPS appears to find better solutions in fewer iterations
                                  than the canonical algorithm, but it is much more dependent on the population
                                  topology."
                                  [Poli et al., 2007] doi: 10.1007/s11721-007-0002-0
                                  [Mendes et al., 2004] doi: 10.1109/TEVC.2004.826074
                                  ri = np.random.uniform(size=nneighbor)
                                  inertia = 0.7289
                                  acc_coeff = phip+phig = 4.1
                                  v = inertia * (v + np.sum(ri*acc_coeff/nneighbor*(p[:]-x)))
                    'nips':       Neighborhood Informed Particle Swarm
                                  'fips' but particles are not informed by all other particles
                                  but only the particles in its neighborhood given by topology.
                                  ri = np.random.uniform(size=nneighbor)
                                  inertia = 0.7289, acc_coeff = phip+phig = 4.1
                                  v = inertia * (v + np.sum(ri*acc_coeff/nneighbor*(p[neighbor[:]]-x)))
        memetic     string
                    Perfom local search at some particle positions
                    (Default: 'no')
                    'no':     No local search.
                    'global': Local search is applied on the overall best position of the swarm.
                    'local':  Local search is applied on each particle with a certain probability.
        nmemetic    integer
                    Do local search after each nmemetic swarm iterations.
                    (Default: 1)
        nlocal      integer
                    Number of local searches.
                    (Default: 5)
        rrwde       float
                    Norm below which local search uses Random Walk with Direction Exploitation (RWDE),
                    otherwise Cognition-Based Local Search (CBLS)
                    (Default: 0.01)
        pls         float or None
                    Probability for each particle to perform local search after nmemetic iterations
                    (Default: 0.2)
                    If None or < 0 then pls will be calculated from the success rate of the last local searches.
                    [Wang et al. 2012] doi: 10.1016/j.ins.2012.02.016
                    Be ngood the actual number of valid local moving steps (at the memetic step before, t-1)
                    and ntotal the total local moving steps carried out during iteration t-1, respectively.
                      rgood = ngood/ntotal
                      delta = 0.1, beta = 0.5, plsmin = 0.1, plsmax = 1.0
                      pls = min(max(beta**np.sign(rgood-delta)*pls, plsmin), plsmax)
        topology    string
                    Neighborhood topologies. These are rather social than geographical topologies.
                    All neighborhoods comprise the current particle as well.
                    [Kennedy & Mendes, 2002] doi: 10.1109/CEC.2002.1004493
                    (Default: 'gbest')
                    'gbest'    Neighborhood is entire swarm.
                    'lbest'    Partciles arranged in a ring, in which each particle communicates with
                               kl particles on each side, i.e. particle i has the neighborhood
                               i-kl, i-kl+1, ..., i, i+1, ..., i+kl-1, i+kl
                               [Mohais et al., 2005] doi: 10.1007/11589990_80
                    'mbest'    Neighborhood is entire part of the swarm on the current MPI process.
                               'mbest' == 'gbest' if run is on one processor.
                               'mbest' gets a more and more local search when mpi processes increase.
                               This is more a gimmick than a real topology.
                    'neumann'  Neighborhood of a point including all points at a Hamming distance of 1.
                               Particles are arranges in a lattice, where each particle interacts with
                               its immediate 4 neighbors to the N, S, E, and W.
                               [Kennedy and Mendes, 2006] doi: 10.1109/TSMCC.2006.875410
                               The von Neumann neighborhood is configured into r rows and c columns,
                               where r is the highest integer less than or equal to sqrt(n) that evenly
                               divides n and c = n / r
                               [Mohais et al., 2005] doi: 10.1007/11589990_80
                    'ring'     'lbest' with kl=1
        kl          integer
                    Neighborhood distance in topology 'lbest'.
                    (Default: 1 = ring)
        verbose     integer
                    Controlling amount of print-out.
                    (Default: 0)
                    0: No print-out.
                    1: Printing convergence criteria, etc.
                    2: Printing after each step.
        includex0   boolean
                    True: include x0 in initial swarm
                    (Default: False)
        seed        int or array_like
                    Seed for numpy's random number generator.
                    (Default: None)
        processes   int
                    The number of processes to use to evaluate objective function and constraints.
                    (Default: 1)
        pout        boolean
                    True: include best per-particle positions and their objective values in output.
                    (Default: False)
        cout        boolean
                    True: include number of function calls in output.
                    (Default: False)
        restart           boolean
                          if True, continue from saved state in restartfile1/2.
                          (Default: False)
        restartfile1/2    string
                          File names for saving current state of PSO.
                          (Default: pso.restart.npz and pso.restart.txt)
                          State will be always written, except if restartfile1=None.
        parameterfile     string
                          Parameter file for executable; must be given if func is name of executable
                          (Default: None)
        parameterwriter   function
                          Python function for writing parameter file if func is name of executable
                          (Default: None)
        objectivefile     string
                          File with objective value from executable; must be given if func is name of executable
                          (Default: None)
        objectivereader   function
                          Python function for reading objective value if func is name of executable
                          (Default: None)
        shell             boolean
                          If True, the specified executable will be executed through the shell.
                          (Default: False)
        debug             boolean
                          If True, model output is displayed for executable.
                          (Default: False)


        Output
        ------
        g           1D-array
                    The swarm's best known position.
        fg          scalar
                    The objective value at g.
        p           ND-array
                    The best known position of each particle.
        fp          1D-array
                    The objective values at each position in p.


        License
        -------
        The original code of Abraham Lee (pyswarm) was published under the BSD license.

        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2013-2016 Abraham Lee, Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  Abraham Lee, 2013 - https://github.com/tisimst/pyswarm
        Modified, MC, Nov 2016 - adapted to JAMS package
                  MC, Nov 2016 - Changed defaults from swarmsize=100, omega=0.5, phip=0.5, phig=0.5, maxn=100
                               - Include vmax for original: clip(v, vmin, vmax)
                               - Sobol sequences and latin hypercube sampling for initial swarm positions
                               - Different PSO algorithms: original, decreasing inertia weights, constricted
                                 fully informed
                               - Stop if function below minobj
                               - debug -> verbose, maxit, cout
                               - neighborhoods
                               - external function - mask, x0, parameterfile, parameterwriter,
                                                     objectivefile, objectivereader, shell, debug
                  MC, Dec 2016 - includex0, restart, mpi, memetic
    """
    # Get MPI communicator
    try:
        from mpi4py import MPI
        comm  = MPI.COMM_WORLD
        csize = comm.Get_size()
        crank = comm.Get_rank()
    except ImportError:
        comm  = None
        csize = 1
        crank = 0
    if csize > 1:
        passrank = crank
    else:
        passrank = None

    dodiv = False

    # Different variabels types (array, float, int, ...) for restart
    if restartfile1 is not None:
        # Only arrays with savez_compressed - restartfile1
        restartarray  = ['x0', 'lb', 'ub', 'mask1', 'mask2', 'x02',
                         'v', 'x', 'fx', 'fs', 'p', 'fp', 'gp', 'fgp',
                         'rs2']
        if dodiv: restartarray.extend(['gx'])
        # Save scalars in simple text file - restartfile2
        restartint    = ['D', 'S', 'it', 'iS', 'crank',
                         'rs3', 'rs4', 'ilocal', 'nlgood', 'nltotal']
        restartfloat  = ['rs5', 'ipls']
        restartbool   = ['maxit']
        restartstring = ['rs1']
        if csize > 1:
            restartfile1 = restartfile1[0:restartfile1.rfind(".")] + '.' + str(crank) + restartfile1[restartfile1.rfind("."):]
            restartfile2 = restartfile2[0:restartfile2.rfind(".")] + '.' + str(crank) + restartfile2[restartfile2.rfind("."):]
        saveargarray = '"'+restartfile1+'"'
        for j in restartarray: saveargarray = saveargarray + ', '+j+'='+j
        saveargint    = ','.join(restartint)
        saveargfloat  = ','.join(restartfloat)
        saveargbool   = ','.join(restartbool)
        saveargstring = ','.join(restartstring)

    # Checks
    # Function
    assert hasattr(func, '__call__') or isinstance(func, (str,list)), 'Invalid function handle or external call.'
    # Bounds
    assert len(lb)==len(ub), 'Lower- and upper-bounds must have the same lengths.'
    lb = np.array(lb)
    ub = np.array(ub)
    assert np.all(ub >= lb), 'All upper-bounds must be greater or equal than lower-bounds.'
    # Mask
    if mask is not None:
        assert len(mask)==len(ub), 'Mask and bounds must have the same lengths.'
        if not np.all(mask):
            assert len(mask)==len(x0), 'Mask and x0 must have the same lengths.'
    # Initialisation keyword
    inits = ['random', 'lhs', 'sobol']
    assert init.lower() in inits, 'Initialisation {:} not in {:}'.format(init, inits)
    if init.lower() == 'sobol':
        assert len(lb) <= 40, "Sobol' sequences only work up to 40 dimensions."
    # Strategy keyword
    ptypes = ['original', 'inertia', 'canonical', 'fips', 'nips']
    assert strategy.lower() in ptypes, 'PSO implementation {:} not in {:}'.format(strategy, ptypes)
    # Topology keyword
    ttypes = ['gbest', 'lbest', 'mbest', 'neumann', 'ring']
    assert topology.lower() in ttypes, 'Topology {:} not in {:}'.format(topology, ttypes)
    # Mmemetic keyword
    mtypes = ['no', 'global', 'local']
    assert memetic.lower() in mtypes, 'Memetic implementation {:} not in {:}'.format(memetic, mtypes)
    # Parameterfile etc. keywords if func is name of executable
    if isinstance(func, (str,list)):
        if parameterfile is None:
            raise IOError('parameterfile must be given if func is name of executable.')
        else:
            if csize > 1:
                parameterfile1 = parameterfile + '.' + str(crank)
            else:
                parameterfile1 = parameterfile
        if parameterwriter is None:
            raise IOError('parameterwrite must be given if func is name of executable.')
        if objectivefile is None:
            raise IOError('objectivefile must be given if func is name of executable.')
        else:
            if csize > 1:
                objectivefile1 = objectivefile + '.' + str(crank)
            else:
                objectivefile1 = objectivefile
        if objectivereader is None:
            raise IOError('objectivereader must be given if func is name of executable.')

    # Set defaults per strategy
    if strategy.lower() == 'original':    # Kennedy & Eberhart, 2001
        if inertia is None: inertia=0.5
        if phip  is None: phip=2.
        if phig  is None: phig=2.
    elif strategy.lower() == 'inertia':   # Shi & Eberhart (1998)
        imax = 0.9
        imin = 0.4
        if phip is None: phip=2.
        if phig is None: phig=2.
    elif strategy.lower() == 'canonical': # Clerc & Kennedy (2000)
        if inertia is None: inertia=0.7289
        if phip  is None: phip=2.05
        if phig  is None: phig=2.05
    elif strategy.lower() == 'fips':      # Mendes & Kennedy (2004)
        if inertia is None: inertia=0.7289
        if phip  is None: phip=2.05
        if phig  is None: phig=2.05
    elif strategy.lower() == 'nips':
        if inertia is None: inertia=0.7289
        if phip  is None: phip=2.05
        if phig  is None: phig=2.05

    if not restart:
        # Problem sizes
        D = len(lb) # dimension of each particle
        if swarmsize is None:
            S = max(min(3*D,40),10)
            S = max(min(S//csize*csize, (40//csize+1)*csize), (10//csize+1)*csize)
        else:
            S = swarmsize

        # Local swarmsize
        if S % csize != 0:
            raise ValueError("Swarmsize "+str(S)+" must be multiple of number of processes "+str(csize)+".")
        iS = S//csize # local swarmsize

        # Initialise 1D mask
        if mask is not None:
            mask1 = mask
        else:
            mask1 = np.ones(D, dtype=np.bool)

        # Partialise objective function
        if isinstance(func, (str,list)):
            obj = partial(_ext_obj_wrapper, func, lb, ub, mask1,
                          parameterfile1, parameterwriter, objectivefile1, objectivereader, shell, debug, passrank)
        else:
            obj = partial(_obj_wrapper, func, arg, kwarg)

        # Check for constraint function(s) and partialise them
        if f_ieqcons is None:
            if ieqcons is None:
                if (verbose>=1) and (crank == 0):
                    print('No constraints given.')
                cons = _cons_none_wrapper
            else:
                if (verbose>=1) and (crank == 0):
                    print('Converting ieqcons to a single constraint function.')
                cons = partial(_cons_ieqcons_wrapper, ieqcons, arg, kwarg)
        else:
            if (verbose>=1) and (crank == 0):
                print('Single constraint function given in f_ieqcons.')
            cons = partial(_cons_f_ieqcons_wrapper, f_ieqcons, arg, kwarg)
        is_feasible = partial(_is_feasible_wrapper, cons)

        # Initialize the multiprocessing module if necessary
        if processes > 1:
            import multiprocessing
            mp_pool = multiprocessing.Pool(processes)

        # 2D mask
        mask2 = np.tile(mask1,iS).reshape((iS,D))
        x02   = np.tile(x0,iS).reshape((iS,D))

        # Deal with NaN and Inf
        large = huge / S

        # Seed random number generator
        if crank == 0:
            np.random.seed(seed=seed)

        # Initialize the particle swarm
        # current particle positions
        # current particle velocities
        if init.lower() == 'random':
            # Random numbers only on rank 0 for reproducible results
            if crank == 0:
                rand = np.random.uniform(size=(2*S,D))
            else:
                rand = np.empty((2*S,D), dtype=np.float)
            # Scatter has different ordering than needed for reproducible results
            # Do it manually
            if csize > 1:
                comm.Bcast(rand, root=0)
            x = rand[crank*iS:crank*iS+iS,:]
            v = rand[S+crank*iS:S+crank*iS+iS,:]
        elif init.lower() == 'sobol':
            import sobol
            nskip = D*S + crank*D*iS
            x = sobol.i4_sobol_generate(D,iS,nskip).transpose()
            nskip = 2*D*S + crank*D*iS
            v = sobol.i4_sobol_generate(D,iS,nskip).transpose()
        elif init.lower() == 'lhs':
            x = np.empty((iS,D), dtype=np.float64)
            v = np.empty((iS,D), dtype=np.float64)
            if crank == 0:
                import scipy.stats as stats
                from pyjams.jams.lhs import lhs
                dist = [stats.uniform for i in range(D)]
                pars = [(0,1) for i in range(D)]
                gx   = lhs(dist, pars, S).transpose()
                gv   = lhs(dist, pars, S).transpose()
            else:
                gx = np.empty((S,D), dtype=np.float64)
                gv = np.empty((S,D), dtype=np.float64)
            if csize > 1:
                comm.Scatter([gx, MPI.DOUBLE], [x, MPI.DOUBLE])
                comm.Scatter([gv, MPI.DOUBLE], [v, MPI.DOUBLE])
            else:
                x = gx
                v = gv
        fx  = np.ones(iS)*large                            # local current particles function values
        if dodiv: gx  = np.empty((iS,D), dtype=np.float64) # global current individual particle positions
        fs  = np.zeros(iS, dtype=bool)                     # current combined feasibility for each local particle
        p   = np.ones((iS,D), dtype=np.float64)*large      # local particles individual best positions
        fp  = np.ones(iS, dtype=np.float64)*large          # local particles individual best function values
        gp  = np.ones((S,D), dtype=np.float64)*large       # global particles individual best positions
        fgp = np.ones(S, dtype=np.float64)*large           # global particles individual best function values

        # Maximum velocity
        vmax = np.abs(ub - lb)
        vmin = -vmax

        # Initialize particle positions and velocities
        v = range012range(v,vmin,vmax)
        x = range012range(x,lb,ub)
        if (crank == 0) and includex0: x[-1,:] = x0
        x = np.where(mask2, x, x02)

        # Calculate first objective and constraints for each particle
        if processes > 1:
            fs = np.array(mp_pool.map(is_feasible, x))
            ii = np.where(fs)[0]
            if ii.size > 0:
                fx[ii] = np.array(mp_pool.map(obj, x[ii,:]))
        else:
            for i in range(iS):
                fs[i] = is_feasible(x[i,:])
                if fs[i]:
                    fx[i] = obj(x[i,:])
        # maximise
        if maxit: fx *= -1.

        # NaN/Inf - ToDo: Check
        large = max(fp.max(), fx[np.isfinite(fx)].max())
        large = 1.1*large if large>0. else 0.9*large
        fx = np.where(np.isfinite(fx), fx, large)

        # print(1, x[np.where(fs)[0],:])
        # Store particles best positions (if constraints are satisfied)
        i_update = (fx < fp) & fs
        if np.any(i_update):
            p[i_update,:] = x[i_update,:].copy()
            fp[i_update]  = fx[i_update]

        # gather local best particles into global best particles
        if csize > 1:
            comm.Allgather([p, MPI.DOUBLE], [gp, MPI.DOUBLE])
            comm.Allgather([fp, MPI.DOUBLE], [fgp, MPI.DOUBLE])
        else:
            gp  = p
            fgp = fp

        if dodiv:
            if csize > 1:
                comm.Allgather([x, MPI.DOUBLE], [gx, MPI.DOUBLE])
            else:
                gx = x
            gmask = np.tile(mask1,S).reshape((S,D))
            divers = diversity(gx, lb, ub, gmask)
            if crank == 0: print(1, divers, fgp.min())

        # Iterate until termination criterion met
        it = 1
        ilocal  = 0
        nlgood  = 0
        nltotal = nlocal
        delta   = 0.1
        beta    = 0.5
        plsmin  = 0.1
        plsmax  = 1.0
        if pls is None:
            dopls = True
        else:
            if pls < 0.:
                dopls = True
            else:
                dopls = False
        if dopls:
            ipls = plsmax
        else:
            ipls = pls

        # save restart
        if restartfile1 is not None:
            if crank == 0:
                rs1, rs2, rs3, rs4, rs5 = np.random.get_state()
            else:
                rs1, rs2, rs3, rs4, rs5 = 'MT19937', np.array(624, dtype=np.uint), 0, 0, 0.
            exec("savez_compressed("+saveargarray+")")
            p2 = open(restartfile2, 'w')
            exec("print("+saveargint+", file=p2)")
            exec("print("+saveargfloat+", file=p2)")
            exec("print("+saveargbool+", file=p2)")
            exec("print("+saveargstring+", file=p2)")
            p2.close()

    else: # if no restart

        # load restart
        p1 = open(restartfile1, 'rb')
        pp = np.load(p1)
        for i in pp.files: exec(i+" = pp['"+i+"']")
        p1.close()
        p2 = open(restartfile2, 'r')
        for i, pp in enumerate(p2.readline().rstrip().split()): exec(restartint[i]+" = int(pp)")
        for i, pp in enumerate(p2.readline().rstrip().split()): exec(restartfloat[i]+" = float(pp)")
        for i, pp in enumerate(p2.readline().rstrip().split()): exec(restartbool[i]+" = bool(strtobool(pp))")
        for i, pp in enumerate(p2.readline().rstrip().split()): exec(restartstring[i]+" = pp")
        p2.close()
        if crank == 0:
            np.random.set_state((rs1, rs2, rs3, rs4, rs5))

        # Partialise objective function
        if isinstance(func, (str,list)):
            obj = partial(_ext_obj_wrapper, func, lb, ub, mask1,
                          parameterfile1, parameterwriter, objectivefile1, objectivereader, shell, debug, passrank)
        else:
            obj = partial(_obj_wrapper, func, arg, kwarg)

        # Check for constraint function(s) and partialise them
        if f_ieqcons is None:
            if ieqcons is None:
                if (verbose>=1) and (crank == 0):
                    print('No constraints given.')
                cons = _cons_none_wrapper
            else:
                if (verbose>=1) and (crank == 0):
                    print('Converting ieqcons to a single constraint function.')
                cons = partial(_cons_ieqcons_wrapper, ieqcons, arg, kwarg)
        else:
            if (verbose>=1) and (crank == 0):
                print('Single constraint function given in f_ieqcons.')
            cons = partial(_cons_f_ieqcons_wrapper, f_ieqcons, arg, kwarg)
        is_feasible = partial(_is_feasible_wrapper, cons)

        if processes > 1:
            import multiprocessing
            mp_pool = multiprocessing.Pool(processes)

    # Iterate swarm
    while it < maxn:
        # Stop if minimum found
        if fgp.min() < minobj:
            if (verbose>=1) and (crank == 0): print('minobj found (1).')
            break
        
        # Memetic PSO
        if (memetic.lower() == 'global') and (it % nmemetic == 0):
            ii0 = fgp.argmin() # gbest
            xl  = gp[ii0,:]
            fxl = fgp[ii0]
            if crank == 0:
                # calc norm in [0,1] space for all parameters
                gp01 = range2range01(gp,lb,ub)
                xl01 = range2range01(xl,lb,ub)
                allnorm = np.array([ np.linalg.norm(gp01[ii,:]-xl01) if ii != ii0 else huge for ii in range(S) ])
                # # stepsize is lower or equal to furthest particle - almost no improvement
                # iib = allnorm.argmax()
                # # stepsize is lower or equal to mean particle distance - quite some steps with improvements
                # iib = closest(allnorm, allnorm.mean())
                # stepsize is lower or equal to closest particle - lots of steps with improvements
                iib = allnorm.argmin()
                dx = np.abs(gp[iib,:]-xl)
                xnew, fxnew, ilgood = rwde(obj, is_feasible,
                                           xl, fxl,
                                           xl-dx, xl+dx,
                                           lb, ub, x0, mask1, maxit, nlocal=nlocal)
                # # take global minimum for x and p
                # xnew, fxnew, ilgood = cbls(obj, is_feasible,
                #                            xl, fxl, # x, f(x)
                #                            xl,      # p - particles best
                #                            xl-dx, xl+dx,
                #                            lb, ub, x0, mask1,
                #                            inertia, phip, maxit, strategy, nlocal=nlocal)
                fxnew  = np.ones(1)*fxnew
            else:
                xnew   = np.empty(D, dtype=np.float64)
                fxnew  = np.empty(1, dtype=np.float64)
            if csize > 1:
                comm.Bcast([xnew, MPI.INT], root=0)
                comm.Bcast([fxnew, MPI.INT], root=0)
            ilocal += nlocal
            if fxnew < fgp[ii0]:
                # if crank==0: print('Mememetic global - iteration, old, new: ', it, fgp[ii0], fxnew[0])
                gp[ii0,:] = xnew
                fgp[ii0]  = fxnew
            if csize > 1:
                comm.Scatter([gp, MPI.DOUBLE], [p, MPI.DOUBLE])
                comm.Scatter([fgp, MPI.DOUBLE], [fp, MPI.DOUBLE])
            else:
                p  = gp
                fp = fgp
            if fgp.min() < minobj:
                if (verbose>=1) and (crank == 0): print('minobj found (2).')
                break

        elif (memetic.lower() == 'local') and (it % nmemetic == 0):
            if crank == 0: # for reproducibility between with and without mpi
                randS = np.random.uniform(size=S)
            else:
                randS = np.empty(S, dtype=np.float)
            if csize > 1:
                comm.Bcast(randS, root=0)
            rS = randS[crank*iS:(crank+1)*iS]
            if dopls:
                if csize > 1:
                    nggood  = comm.allreduce(nlgood, op=MPI.SUM)
                    ngtotal = comm.allreduce(nltotal, op=MPI.SUM)
                else:
                    nggood  = nlgood
                    ngtotal = nltotal
                rgood = float(nggood)/float(max(ngtotal,1))
                ipls  = min(max(beta**np.sign(rgood-delta)*ipls, plsmin), plsmax)
            nlgood  = 0
            nltotal = 0
            for i in range(iS):
                if rS[i] < ipls:
                    xl      = x[i,:]
                    fxl     = fx[i]
                    x01     = range2range01(x,lb,ub) # norm in [0,1] space
                    xl01    = x01[i,:]
                    allnorm = np.array([ np.linalg.norm(x01[ii,:]-xl01) if ii != i else huge for ii in range(iS) ])
                    iib     = allnorm.argmin()       # stepsize is lower or equal to closest particle
                    dx      = np.abs(x[iib,:]-xl)    # stepzise in real parameter range
                    # dx      = np.abs(p[i,:]-xl)      # stepzise in real parameter range
                    norm    = np.linalg.norm(range2range01(p[i,:],lb,ub)-xl01)
                    if norm < rrwde:
                        xnew, fxnew, ilgood = rwde(obj, is_feasible,
                                                   xl, fxl,
                                                   xl-dx, xl+dx,
                                                   lb, ub, x0, mask1, maxit, nlocal=nlocal)
                    else:
                        xnew, fxnew, ilgood = cbls(obj, is_feasible,
                                                   xl, fxl, # x, f(x)
                                                   p[i,:],  # p - particles best
                                                   xl-dx, xl+dx,
                                                   lb, ub, x0, mask1,
                                                   inertia, phip, maxit, strategy, nlocal=nlocal)
                    x[i,:] = xnew
                    fx[i]  = fxnew
                    nlgood  += ilgood
                    nltotal += nlocal
            ilocal += nltotal
            # Store and scatter particles best positions
            i_update = fx < fp
            if np.any(i_update):
                # print('Mememetic local - it, rank, old, new: ', it, crank, fp[i_update], fx[i_update])
                p[i_update,:] = x[i_update,:].copy()
                fp[i_update]  = fx[i_update]
                # gather local best particles into global best particles
            if csize > 1:
                comm.Allgather([p, MPI.DOUBLE], [gp, MPI.DOUBLE])
                comm.Allgather([fp, MPI.DOUBLE], [fgp, MPI.DOUBLE])
            else:
                gp  = p
                fgp = fp
            if fgp.min() < minobj:
                if (verbose>=1) and (crank == 0): print('minobj found (2).')
                break

        # Update neighbors best positions
        g  = np.empty((iS,D), dtype=np.float64) # local best neighbors
        fg = np.empty(iS, dtype=np.float64)
        if topology.lower() == 'mbest':
            i_min = np.argmin(fp) # overall best on MPI process
            ig  = p[i_min,:]
            ifg = fp[i_min]
            for iis in range(iS): g[iis,:] = ig
            fg[:] = ifg
        else:
            if crank == 0:
                gg, fgg = get_best_neighbor(gp, fgp, topology, kl=kl)  # global best neighbors
            else:
                gg  = np.empty((S,D), dtype=np.float64)
                fgg = np.empty(S, dtype=np.float64)
            # Scatter global best neighbors into local best neighbors
            if csize > 1:
                comm.Scatter([gg, MPI.DOUBLE], [g, MPI.DOUBLE])
                comm.Scatter([fgg, MPI.DOUBLE], [fg, MPI.DOUBLE])
            else:
                g  = gg
                fg = fgg

        # Update the particles velocities
        if crank == 0:
            rand = np.random.uniform(size=(2*S+S*S,D))
        else:
            rand = np.empty((2*S+S*S,D), dtype=np.float)
        # Scatter has different ordering than needed for reproducible results
        # Do it manually
        if csize > 1:
            comm.Bcast(rand, root=0)
        rp = rand[crank*iS:crank*iS+iS,:]
        rg = rand[S+crank*iS:S+crank*iS+iS,:]
        if strategy.lower() == 'original':    # Kennedy & Eberhart, 2001
            v = inertia*v + phip*rp*(p - x) + phig*rg*(g - x)
            v = np.clip(v, vmin, vmax)
        elif strategy.lower() == 'inertia':   # Shi & Eberhart (1998)
            inertia = imax - float(it)/float(maxn-1) * (imax-imin)
            v = inertia*v + phip*rp*(p - x) + phig*rg*(g - x)
            v = np.clip(v, vmin, vmax)
        elif strategy.lower() == 'canonical': # Clerc & Kennedy (2000)
            v = inertia * (v + phip*rp*(p - x) + phig*rg*(g - x))
            # limit v also to x-bounds
            # Eberhardt RC & Shi Y (2000)
            # Comparing inertia weights and constriction factors in particle swarm optimization.
            # In: Proceedings of the Congress on Evolutionary Computation (CEC2000), 84-88.
            v = np.clip(v, lb, ub)
        elif strategy.lower() == 'fips':      # Mendes & Kennedy (2004)
            acc_coeff = (phip + phig) / float(S)
            for i in range(iS):
                ri = rand[2*S+crank*iS*S+i*S:2*S+crank*iS*S+(i+1)*S,:]
                v[i,:] = inertia * (v[i,:] + np.sum(ri[:,:]*acc_coeff*(gp[:,:]-x[i,:]), axis=0))
        elif strategy.lower() == 'nips':      # Mendes & Kennedy (2004)
            acc_coeff = (phip + phig) / float(S)
            for i in range(iS):
                ri = rand[2*S+crank*iS*S+i*S:2*S+crank*iS*S+(i+1)*S,:]
                ii = get_neighbor_indeces(crank*iS+i, S, topology, kl=kl)
                v[i,:] = inertia * (v[i,:] + np.sum(ri[ii,:]*acc_coeff*(gp[ii,:]-x[i,:]), axis=0))

        # Update the particles positions
        x = x + v
        x = np.where(mask2, x, x02)

        # Limit to bounds
        x = np.clip(x, lb, ub)

        # Update objectives and constraints
        if processes > 1:
            fs = np.array(mp_pool.map(is_feasible, x))
            ii = np.where(fs)[0]
            if ii.size > 0:
                fx[ii] = np.array(mp_pool.map(obj, x[ii,:]))
                if maxit: fx[ii] *= -1.
        else:
            for i in range(iS):
                fs[i] = is_feasible(x[i,:])
                if fs[i]:
                    fx[i] = obj(x[i,:])
                    if maxit: fx[i] *= -1.

        # NaN/Inf
        large = max(fp.max(), fx[np.isfinite(fx)].max())
        large = 1.1*large if large>0. else 0.9*large
        fx = np.where(np.isfinite(fx), fx, large)

        # print(2, x[np.where(fs)[0],:])
        # Store particles best positions (if constraints are satisfied)
        i_update = (fx < fp) & fs
        if np.any(i_update):
            p[i_update,:] = x[i_update,:].copy()
            fp[i_update]  = fx[i_update]

        # gather local best particles into global best particles
        if csize > 1:
            comm.Allgather([p, MPI.DOUBLE], [gp, MPI.DOUBLE])
            comm.Allgather([fp, MPI.DOUBLE], [fgp, MPI.DOUBLE])
        else:
            gp  = p
            fgp = fp

        if dodiv:
            if csize > 1:
                comm.Allgather([x, MPI.DOUBLE], [gx, MPI.DOUBLE])
            else:
                x = gx
            gmask = np.tile(mask1,S).reshape((S,D))
            divers = diversity(gx, lb, ub, gmask)
            if crank == 0: print(2, divers, fgp.min())

        it += 1

        # save restart
        if restartfile1 is not None:
            if crank == 0:
                rs1, rs2, rs3, rs4, rs5 = np.random.get_state()
            else:
                rs1, rs2, rs3, rs4, rs5 = 'MT19937', np.array(624, dtype=np.uint), 0, 0, 0.
            exec("savez_compressed("+saveargarray+")")
            p2 = open(restartfile2, 'w')
            exec("print("+saveargint+", file=p2)")
            exec("print("+saveargfloat+", file=p2)")
            exec("print("+saveargbool+", file=p2)")
            exec("print("+saveargstring+", file=p2)")
            p2.close()
    # end of swarm iteration: while it < maxn:

    if (it == maxn) and (verbose>=1) and (crank == 0): print('Maximum iterations reached --> {:}.'.format(maxn))

    # global best
    i_min = np.argmin(fgp)
    bestx = gp[i_min,:].copy()
    bestf = fgp[i_min]

    if maxit:
        bestf *= -1.
        fgp   *= -1.

    if not any(map(is_feasible, gp)) and (crank == 0): print("PSO could not find any feasible point in the search space.")

    # write parameter file with best parameters
    if crank == 0:
        if isinstance(func, (str,list)):
            parameterwriter(parameterfile, bestx, lb, ub, mask1)

    if csize > 1:
        tlocal = comm.allreduce(ilocal, op=MPI.SUM)
    else:
        tlocal = ilocal

    out = [bestx, bestf]
    if pout:
        out += [p, fp]
    if cout:
        out += [it*S+tlocal]

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # try:
    #     from mpi4py import MPI
    #     comm  = MPI.COMM_WORLD
    #     csize = comm.Get_size()
    #     crank = comm.Get_rank()
    # except ImportError:
    #     comm  = None
    #     csize = 1
    #     crank = 0

    # algo      = 'canonical'
    # init      = 'lhs'
    # swarmsize = 40
    # maxn      = 250
    # topology  = 'neumann'
    # memetic   = 'local'
    # nmemetic  = 2

    # if swarmsize % csize != 0:
    #     raise ValueError("Swarmsize "+str(swarmsize)+" must be multiple of number of processes "+str(csize)+".")

    # from jams.functions import ackley, griewank, goldstein_price, rastrigin, rosenbrock, six_hump_camelback
    # '''
    # This is the Ackley Function
    # Global Optimum (n>=2): 0.0 at origin
    # '''
    # npara = 10
    # lb = -10*np.ones(npara)
    # ub = 10*np.ones(npara)
    # x0 = np.zeros(npara)
    # bestx, bestf, nrun = pso(ackley, x0, lb, ub, processes=1,
    #                          init=init, strategy=algo, topology=topology, verbose=0,
    #                          memetic=memetic, nmemetic=nmemetic,
    #                          swarmsize=swarmsize, maxn=maxn, restartfile1=None, cout=True)
    # if crank == 0: print('Ackley 0 at origin ', nrun, bestx, bestf)
    # '''
    #     This is the Griewank Function (2-D or 10-D)
    #     Bound: X(i)=[-600,600], for i=1,2,...,10  !for visualization only 2!
    #        Global Optimum: 0, at origin
    # '''
    # npara = 10
    # lb = -600*np.ones(npara)
    # ub = 600*np.ones(npara)
    # x0 = np.zeros(npara)
    # bestx, bestf, nrun = pso(griewank, x0, lb, ub, processes=1,
    #                          init=init, strategy=algo, topology=topology, verbose=0,
    #                          memetic=memetic, nmemetic=nmemetic,
    #                          swarmsize=swarmsize, maxn=maxn, restartfile1=None, cout=True)
    # if crank == 0: print('Griewank 0 at origin ', nrun, bestx, bestf)
    # '''
    # This is the Goldstein-Price Function
    # Bound X1=[-2,2], X2=[-2,2]
    # Global Optimum: 3.0,(0.0,-1.0)
    # '''
    # npara = 2
    # lb = -2*np.ones(npara)
    # ub = 2*np.ones(npara)
    # x0 = np.zeros(npara)
    # bestx, bestf, nrun = pso(goldstein_price, x0, lb, ub, processes=4, minobj=3.+1e-8,
    #                          init=init, strategy=algo, topology=topology, verbose=0,
    #                          memetic=memetic, nmemetic=nmemetic,
    #                          swarmsize=swarmsize, maxn=maxn, restartfile1=None, cout=True)
    # if crank == 0: print('Goldstein 3 at (0,-1) ', nrun, bestx, bestf)
    # '''
    # This is the Rastrigin Function
    # Bound: X1=[-1,1], X2=[-1,1]
    # Global Optimum: -2, (0,0)
    # '''
    # npara = 2
    # lb = -1*np.ones(npara)
    # ub = 1*np.ones(npara)
    # x0 = np.zeros(npara)
    # bestx, bestf, nrun = pso(rastrigin, x0, lb, ub, processes=4, minobj=-2.+1e-8,
    #                          init=init, strategy=algo, topology=topology, verbose=0,
    #                          memetic=memetic, nmemetic=nmemetic,
    #                          swarmsize=swarmsize, maxn=maxn, restartfile1=None, cout=True)
    # if crank == 0: print('Rastrigin -2 at origin ', nrun, bestx, bestf)
    # '''
    # This is the Rosenbrock Function
    # Bound: X1=[-5,5], X2=[-2,8]; Global Optimum: 0,(1,1)
    #        lb=[-5 -5]; ub=[5 5]; x0=[1 1];
    # '''
    # npara = 2
    # lb = -2*np.ones(npara)
    # ub = 5*np.ones(npara)
    # x0 = np.zeros(npara)
    # bestx, bestf, nrun = pso(rosenbrock, x0, lb, ub, processes=4,
    #                          init=init, strategy=algo, topology=topology, verbose=0,
    #                          memetic=memetic, nmemetic=nmemetic,
    #                          swarmsize=swarmsize, maxn=maxn, restartfile1=None, cout=True)
    # if crank == 0: print('Rosenbrock 0 at (1,1) ', nrun, bestx, bestf)
    # '''
    # This is the Six-hump Camelback Function.
    # Bound: X1=[-5,5], X2=[-5,5]
    # True Optima: -1.031628453489877, (-0.08983,0.7126), (0.08983,-0.7126)
    # '''
    # npara = 2
    # lb = -5*np.ones(npara)
    # ub = 5*np.ones(npara)
    # x0 = np.zeros(npara)
    # bestx, bestf, nrun = pso(six_hump_camelback, x0, lb, ub, processes=4, minobj=-1.031628453489877+1e-8,
    #                          init=init, strategy=algo, topology=topology, verbose=0,
    #                          memetic=memetic, nmemetic=nmemetic,
    #                          swarmsize=swarmsize, maxn=maxn, restartfile1=None, cout=True)
    # if crank == 0: print('Six_hump_camelback -1.03... +-(-0.08983,0.7126) ', nrun, bestx, bestf)

    # # Restart
    # algo      = 'fips'
    # init      = 'lhs'
    # swarmsize = 40
    # maxn      = 250
    # topology  = 'neumann'
    # memetic   = 'no' # 'global'
    # nmemetic  = 1
    # if swarmsize % csize != 0:
    #     raise ValueError("Swarmsize "+str(swarmsize)+" must be multiple of number of processes "+str(csize)+".")
    # from jams.functions import rosenbrock
    # npara = 2
    # lb = -2*np.ones(npara)
    # ub = 5*np.ones(npara)
    # x0 = np.zeros(npara)
    # seed = 1234
    # bestx, bestf = pso(rosenbrock, x0, lb, ub, processes=4,
    #                    init=init, strategy=algo, topology=topology, verbose=0,
    #                    memetic=memetic, nmemetic=nmemetic,
    #                    swarmsize=swarmsize, maxn=maxn, restartfile1=None,
    #                    seed=seed, restart=False)
    # if crank == 0: print('Rosenbrock Reference - ', bestx, bestf)
    # bestx, bestf = pso(rosenbrock, x0, lb, ub, processes=4,
    #                    init=init, strategy=algo, topology=topology, verbose=0,
    #                    memetic=memetic, nmemetic=nmemetic,
    #                    swarmsize=swarmsize, maxn=maxn//2,
    #                    seed=seed, restart=False)
    # if crank == 0: print('Rosenbrock Restart 1 - ', bestx, bestf)
    # bestx, bestf = pso(rosenbrock, x0, lb, ub, processes=4,
    #                    init=init, strategy=algo, topology=topology, verbose=0,
    #                    memetic=memetic, nmemetic=nmemetic,
    #                    swarmsize=swarmsize, maxn=maxn,
    #                    seed=seed, restart=True)
    # if crank == 0: print('Rosenbrock Restart 2 - ', bestx, bestf)

    # # Constraints
    # algo      = 'canonical'
    # init      = 'lhs'
    # swarmsize = 40
    # maxn      = 250
    # topology  = 'gbest'
    # memetic   = 'no' # 'global'
    # nmemetic  = 1
    # if swarmsize % csize != 0:
    #     raise ValueError("Swarmsize "+str(swarmsize)+" must be multiple of number of processes "+str(csize)+".")
    # from jams.functions import rosenbrock
    # npara = 2
    # lb = -2*np.ones(npara)
    # ub = 5*np.ones(npara)
    # x0 = np.zeros(npara)
    # seed = 1234
    # def fcon(x): # feasible = np.all(fcon(x)>=0.)
    #     return np.sign(x)
    # bestx, bestf = pso(rosenbrock, x0, lb, ub, processes=1, f_ieqcons=fcon,
    #                    init=init, strategy=algo, topology=topology, verbose=2,
    #                    memetic=memetic, nmemetic=nmemetic,
    #                    swarmsize=swarmsize, maxn=maxn, restartfile1=None,
    #                    seed=seed, restart=False)
    # if crank == 0: print('Rosenbrock constraint - ', bestx, bestf)

    # # Clean up doctest
    # import os
    # os.remove('pso.restart.npz')
    # os.remove('pso.restart.txt')
